"""
Functions for writing ImageJ Scripts to provided directory
"""

# Imports
# Standard Library Imports
from __future__ import annotations
import importlib.resources
import os
import pathlib
from typing import Union, Literal

# External Imports

# Local Imports


def write_script(
    scriptname: Literal[
        "openformasking",
        "openformaskingsingle",
        "savefishmask",
        "savefishnullmask",
        "macroize",
    ],
    outdir: Union[str, os.PathLike],
) -> None:
    outdir = pathlib.Path(outdir)  # Convert to a pathlib Path
    scriptname = (
        scriptname.lower()
    )  # Convert to lower to avoid issues of capitalization
    scriptname = scriptname.split(".")[
        0
    ]  # If an extension is provided, just take the script name
    if scriptname == "openformasking":
        target_file = outdir / "OpenForMasking.ijm"
        with importlib.resources.open_text(
            "pepitatools.data", "OpenForMasking.ijm"
        ) as script:
            with open(target_file, "w") as f:
                f.write(script.read())
    elif scriptname == "openformaskingsingle":
        target_file = outdir / "OpenForMaskingSingle.ijm"
        with importlib.resources.open_text(
            "pepitatools.data", "OpenForMaskingSingle.ijm"
        ) as script:
            with open(target_file, "w") as f:
                f.write(script.read())
    elif scriptname == "savefishmask":
        target_file = outdir / "SaveFishMask.ijm"
        with importlib.resources.open_text(
            "pepitatools.data", "SaveFishMask.ijm"
        ) as script:
            with open(target_file, "w") as f:
                f.write(script.read())
    elif scriptname == "savefishnullmask":
        target_file = outdir / "SaveFishNullMask.ijm"
        with importlib.resources.open_text(
            "pepitatools.data", "SaveFishNullMask.ijm"
        ) as script:
            with open(target_file, "w") as f:
                f.write(script.read())
    elif scriptname == "macroize":
        target_file = outdir / "macroize.sh"
        with importlib.resources.open_text("pepitatools.data", "macroize.sh") as script:
            with open(target_file, "w") as f:
                f.write(script.read())



"""
Calculate dose response
"""

# Imports
# Standard Library Imports
import csv
import importlib.resources
import os
import os.path
from time import time
import warnings

# External Imports
import numpy as np
import pandas as pd
from matplotlib.ticker import PercentFormatter
import matplotlib.pyplot as plt
import scipy.optimize
import seaborn as sns

# Local Imports
from . import utils
from .configuration import get_config_setting


_neo_model = None


class Model:
    def __init__(self, xs, ys, cocktail, E_0=100, E_max=None):
        self.cocktail = cocktail
        self.combo = len(cocktail.drugs) > 1
        self.xs = xs
        self.ys = ys
        self.E_0 = E_0
        self.E_max = E_max

        self.equation = lambda xs, b, c, e: log_logistic_model(xs, b, c, E_0, e)
        if ys and len(ys) >= 4:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", RuntimeWarning)
                warnings.simplefilter("ignore", scipy.optimize.OptimizeWarning)
                fits = {}
                for method in ("lm", "trf", "dogbox"):
                    try:
                        popt, _ = scipy.optimize.curve_fit(
                            self.equation, self.xs, self.ys, method=method
                        )
                        sse = sum((self.equation(xs, *popt) - ys) ** 2)
                        fits[sse] = popt
                    except Exception as _:  # Bare except?
                        pass
                if not fits:
                    self.b, self.c, self.e = None, None, None
                else:
                    self.b, self.c, self.e = fits[min(fits)]
        else:
            self.b, self.c, self.e = None, None, None

    # The __repr__ is redefined below
    # def __repr__(self):
    #     return "{}({})".format(self.__class__.__name__, self.cocktail)

    def chart(
        self,
        close=True,
        color="darkgrey",
        datapoints=None,
        label=True,
        name=None,
        scale=None,
    ):
        if scale is None:
            scale = [self.E_max, self.E_0]

        ys = [(value - scale[0]) / (scale[1] - scale[0]) for value in self.ys]

        with sns.axes_style(style="darkgrid"):
            plt.scatter(self.xs, ys, color="black", marker="_", s=256)

            ax = plt.gca()
            # ax.set_ylim(bottom=0)
            ax.yaxis.set_major_formatter(PercentFormatter(1))

            if datapoints:
                scores = [value for values in datapoints.values() for value in values]
                scores = [
                    (value - scale[0]) / (scale[1] - scale[0]) for value in scores
                ]

                data = pd.DataFrame(
                    {
                        "brightness": scores,
                        "concentration": [
                            float(soln)
                            for soln, values in datapoints.items()
                            for _ in values
                        ],
                    }
                )
                sns.scatterplot(
                    x="concentration",
                    y="brightness",
                    data=data,
                    color="black",
                    label="Measurements",
                    marker=".",
                    s=64,
                )

            if self.b:
                line_xs = np.linspace(0, float(max(self.xs)), 100)
                line_ys = [
                    (value - scale[0]) / (scale[1] - scale[0])
                    for value in self.get_ys(line_xs)
                ]
                sns.lineplot(x=line_xs, y=line_ys, label="Model")

                ec50 = self.effective_concentration(0.5)

                ec50_label = (
                    f"{ec50:.1f}{self.get_x_units()}" if not np.isnan(ec50) else "N/A"
                )

                plt.scatter(
                    ec50,
                    0.04,
                    color="black",
                    label=f"EC50={ec50_label}",
                    marker="|",
                    s=128,
                )

        if label:
            plt.title(f"{self.get_condition()} Dose-Response Curve")
            plt.xlabel(f"{self.get_condition()} Dose ({self.get_x_units()})")
            plt.ylabel("Remaining Brightness")
            plt.legend()

        if close:
            plt.tight_layout()
            uniq_str = str(int(time() * 1000) % 1_620_000_000_000)
            if not name:
                name = self.get_condition()
            log_dir = f'{get_config_setting("log_dir")}/dose_response'
            os.makedirs(log_dir, exist_ok=True)
            plt.savefig(os.path.join(log_dir, f"{name}_{uniq_str}.png"))
            plt.close()
            plt.clf()

    # pct_survival = (f(x) - min) / (max - min)
    # f(x) = c + (d - c) / (1 + (x / e)**b)
    # yields x
    def effective_concentration(self, pct_inhibition, silent=False):
        if pct_inhibition <= 0 or pct_inhibition >= 1:
            raise RuntimeError("Inhibition level must be between 0 and 1")

        if not self.b:
            return np.nan

        b = self.b
        c = self.c
        d = self.E_0
        e = self.e
        max_ = self.E_0
        min_ = self.get_absolute_E_max()
        pct_survival = 1 - pct_inhibition

        pct_pts_above_E_max = pct_survival * (max_ - min_) + min_ - c

        if pct_pts_above_E_max <= 0:
            if not silent:
                print(
                    f"WARN: {self.get_condition()} EC_{int(pct_inhibition*100)} is unreachable"
                )
            return np.nan

        return e * ((d - c) / pct_pts_above_E_max - 1) ** (1 / b)

    def get_absolute_E_max(self):
        return self.E_max if self.E_max is not None else self.c

    def get_condition(self):
        return str(self.cocktail)

    def get_condition_E_max(self):
        return self.c

    def get_intersection(self, other, guess, ratio):
        if not isinstance(other, Model):
            return np.nan

        other_adj = Model(
            np.array(other.xs) * ratio, other.ys, other.cocktail, other.E_0, other.E_max
        )

        def f_intersection_equals_zero(xs):
            return np.array(
                self.equation(xs, self.b, self.c, self.e)
                - other_adj.equation(xs, other_adj.b, other_adj.c, other_adj.e),
                dtype=np.float64,
            )

        plot_func(
            self.xs,
            f_intersection_equals_zero,
            f"{self.cocktail} - {other_adj.cocktail}",
            f"a_b_intersection@{ratio}",
            f"{self.cocktail} ({self.get_x_units()})",
        )

        return get_intersection(
            lambda xs: self.equation(xs, self.b, self.c, self.e),
            lambda xs: other_adj.equation(xs, other_adj.b, other_adj.c, other_adj.e),
            guess,
        )

    # pct_survival = (f(x) - min) / (max - min)
    def get_pct_survival(self, xs=None, ys=None):
        if xs is None and ys is None:
            raise ValueError("One of xs or ys is required")

        max_ = self.E_0
        min_ = self.get_absolute_E_max()

        if not ys:
            ys = self.get_ys(xs)

        return (ys - min_) / (max_ - min_)

    def get_x_units(self):
        return self.xs[-1].get_units()

    def get_ys(self, xs):
        return self.equation(xs, self.b, self.c, self.e)

    def pivot(self):
        xs = [x.reverse() for x in self.xs]
        return Model(xs, self.ys, xs[-1].get_cocktail(), self.E_0, self.E_max)

    def __repr__(self):
        return str(self.__dict__)


def analyze_checkerboard(
    model_a,
    model_b,
    models_combo,
    method="interpolation",
    file_name_context=None,
    effect_level=0.5,
):
    file_name_context1 = ""
    file_name_context2 = ""
    if file_name_context:
        file_name_context1 = "(" + file_name_context + ")"
        file_name_context2 = file_name_context + "_"

    if method == "interpolation":
        fig = plt.figure(figsize=(12, 8), dpi=100)
        ax = fig.add_subplot(1, 1, 1)
        ax.margins(0.006)

        # Swapped is never used
        # swapped = False

        if model_a.c < model_b.c:
            model_a, model_b = model_b, model_a
            models_combo = [model_combo.pivot() for model_combo in models_combo]
        # swapped = True

        effect_pretty = f"{(effect_level * 100):.0f}"
        ec_a = model_a.effective_concentration(effect_level)
        ec_b = model_b.effective_concentration(effect_level)
        print(
            f"{model_a.cocktail} EC{effect_pretty} = {ec_a:.1f}{model_a.get_x_units()}"
        )
        print(
            f"{model_b.cocktail} EC{effect_pretty} = {ec_b:.1f}{model_b.get_x_units()}"
        )

        plt.title(
            (
                f"{model_a.cocktail}+{model_b.cocktail} "
                f"Predicted vs. Observed $EC_{{{effect_pretty}}}$ Isobole"
            )
        )
        plt.scatter(
            ec_a,
            0,
            color="black",
            label=f"Interpolated Equipotent Single Dose, $EC_{{{effect_pretty}}}$",
            s=16,
        )
        plt.scatter(0, ec_b, color="black", s=16)

        fics, max_x, max_y = [], 1, 1

        for model_combo in models_combo:
            if not model_combo.b:  # some combos will not have enough datapoints: skip
                continue

            ec_combo = model_combo.effective_concentration(effect_level)
            ec_combo_a = ec_combo * model_combo.cocktail.ratio.to_proportion()
            ec_combo_b = (
                ec_combo * model_combo.cocktail.ratio.reciprocal().to_proportion()
            )

            ec_combo_theor = get_combo_additive_expectation(
                effect_level,
                model_a,
                model_b,
                model_combo,
                model_combo.cocktail.ratio,
                plot=False,
            )

            color = "tab:red" if ec_combo > ec_combo_theor else "tab:green"
            plt.scatter(ec_combo_a, ec_combo_b, color=color, s=16)

            fic = get_combo_FIC(
                effect_level, model_a, model_b, model_combo, model_combo.cocktail.ratio
            )
            fics.append(fic)

            offset_x = (
                max(model_a.xs) / 64
            )  # arbitrary adjustments to put text in nice location
            offset_y = max(model_b.xs) / 128
            ax.annotate(
                f"$FIC_{{{effect_pretty}}}={fic:.2f}$",
                xy=(ec_combo_a + offset_x, ec_combo_b - offset_y),
                textcoords="data",
            )

            print(
                (
                    f"{model_combo.cocktail} EC{effect_pretty} = "
                    f"{ec_combo_a:.1f}{model_a.get_x_units()} + {ec_combo_b:.1f}{model_b.get_x_units()}"
                    f"; FIC = {fic:.2f}; ({len(model_combo.ys)} datapoints)"
                )
            )

            max_x = max(ec_a + offset_x, ec_combo_a + offset_x, max_x)
            max_y = max(ec_b + offset_y, ec_combo_b + offset_y, max_y)

        ax.plot(
            [],
            [],
            color="tab:green",
            fillstyle="left",
            label=f"Interpolated Equipotent Combo Doses, $EC_{{{effect_pretty}}}$",
            linestyle="none",
            marker=".",
            markerfacecoloralt="tab:red",
            markeredgewidth=0,
            markersize=12,
        )

        combined_fic = utils.geometric_mean(fics)

        print(
            "Combined FIC for the whole checkerboard (by interpolation analysis):",
            f"{combined_fic:.2f}",
        )

        inhibition_max_a = 1 - model_a.get_pct_survival(ys=model_a.c)
        inhibition_max_b = 1 - model_a.get_pct_survival(ys=model_b.c)

        def f_isobole(ec_combo_a):
            return do_additive_isobole(
                ec_combo_a,
                model_a.e,
                model_b.e,
                inhibition_max_a,
                inhibition_max_b,
                ec_b,
                model_b.b,
                model_a.b,
            )

        plot_func(
            model_a.xs,
            f_isobole,
            f"{model_a.cocktail}+{model_b.cocktail} Predicted $EC_{{{effect_pretty}}}$ Isobole",
            None,
            close=False,
            color="tab:gray",
            max_x=ec_a,
            max_y=ec_b,
            min_x=0,
            min_y=0,
            x_label=f"{model_a.cocktail} Dose ({model_a.get_x_units()})",
            y_label=f"{model_b.cocktail} Dose ({model_b.get_x_units()})",
        )

        plt.xlim(right=max_x)
        plt.ylim(top=max_y)
        plt.tight_layout()
        uniq_str = str(int(time() * 1000) % 1_620_000_000_000)
        log_dir = f'{get_config_setting("log_dir")}/dose_response'
        os.makedirs(log_dir, exist_ok=True)
        plt.savefig(
            os.path.join(
                log_dir,
                f"{model_a.cocktail}+{model_b.cocktail}_isoboles_{uniq_str}.png",
            )
        )
        plt.close()
        plt.clf()
    elif method == "Bliss":
        # following Bansal 2014, https://doi.org/10.1038/nbt.3052

        label_a = f"{model_a.cocktail} Concentration ({model_a.get_x_units()})"
        label_b = f"{model_b.cocktail} Concentration ({model_b.get_x_units()})"

        data_dict = {}
        data_dict[label_a] = [float(x) for x in model_a.xs]
        data_dict[label_b] = [0] * len(model_a.xs)
        data_dict["Bliss Interaction"] = [np.nan for y in model_a.ys]
        data_dict[label_a] = np.append(data_dict[label_a], [0] * len(model_b.xs))
        data_dict[label_b] = np.append(
            data_dict[label_b], [float(x) for x in model_b.xs]
        )
        data_dict["Bliss Interaction"] = np.append(
            data_dict["Bliss Interaction"], [np.nan for y in model_b.ys]
        )

        for model_combo in models_combo:
            xs = np.array([float(x) for x in model_combo.xs])
            data_dict[label_a] = np.append(
                data_dict[label_a], xs * model_combo.cocktail.ratio.to_proportion()
            )
            data_dict[label_b] = np.append(
                data_dict[label_b],
                xs * model_combo.cocktail.ratio.reciprocal().to_proportion(),
            )
            data_dict["Bliss Interaction"] = np.append(
                data_dict["Bliss Interaction"],
                [
                    get_bliss_ixn(x, y, model_a, model_b, model_combo)
                    for x, y in zip(model_combo.xs, model_combo.ys)
                ],
            )

        data = pd.DataFrame(data_dict)
        data = data.pivot_table(
            index=label_a,
            columns=label_b,
            values="Bliss Interaction",
            aggfunc="mean",
            dropna=False,
        )

        fig = plt.figure(figsize=(12, 8), dpi=100)
        ax = sns.heatmap(
            data,
            vmin=-1,
            vmax=1,
            cmap="vlag_r",
            center=0,
            annot=True,
            fmt=".2f",
            linewidths=2,
            square=True,
            cbar_kws={
                "extend": "both",
                "label": "Excess Over Bliss",
                # 'location': 'bottom',
                # 'shrink': 0.5,
                "ticks": [-1, 0, 1],
            },
        )
        ax.collections[0].colorbar.set_ticklabels(
            ["-1 (Antagonism)", "0 (Noninteraction)", "+1 (Synergy)"]
        )
        ax.invert_yaxis()
        plt.title(
            f"{model_a.get_condition()} vs. {model_b.get_condition()}: Bliss Ixn "
            + f"{file_name_context1}"
        )
        plt.tight_layout()
        uniq_str = str(int(time() * 1000) % 1_620_000_000_000)
        log_dir = f'{get_config_setting("log_dir")}/dose_response'
        os.makedirs(log_dir, exist_ok=True)
        plt.savefig(
            f"{log_dir}/{model_a.get_condition()}-{model_b.get_condition()}_{file_name_context2}"
            + f"bliss_{uniq_str}.png"
        )
        plt.clf()
    elif method == "Loewe":
        pass
    else:
        raise ValueError('`method` must be one of "interpolation", "Bliss", or "Loewe"')


def analyze_diamond(model_a, model_b, model_combo):
    if model_a.c is None or model_b.c is None:
        print("Cannot analyze diamond due to failed dose-response curve fit")
        return None, 0, 0

    # print significant statistics
    if model_a.c < model_b.c:
        model_a, model_b = model_b, model_a
        model_combo = model_combo.pivot()

    ax = plt.gca()

    def f_diagonal(ec_combo_a):
        return ec_combo_a / model_combo.cocktail.ratio

    plot_func(
        model_a.xs,
        f_diagonal,
        "Diagonal",
        None,
        close=False,
        color="lightgrey",
        max_x=max(model_a.xs),
        min_x=0,
        x_label=f"{model_a.cocktail} Dose ({model_a.get_x_units()})",
        y_label=f"{model_b.cocktail} Dose ({model_b.get_x_units()})",
    )

    max_x, max_y = 0, 0

    e_experimental = 0.5
    concentration_a = model_a.effective_concentration(e_experimental)
    concentration_b = model_b.effective_concentration(e_experimental)

    print(
        f"Intercept 1: ({concentration_a}{model_a.get_x_units()} {model_a.cocktail}, 0)"
    )
    print(
        f"Intercept 2: (0, {concentration_b}{model_b.get_x_units()} {model_b.cocktail})"
    )
    plt.scatter(
        concentration_a,
        0,
        color="black",
        label=f"Equipotent Single Dose, $EC_{{{(e_experimental * 100):.0f}}}$",
        s=16,
    )
    plt.scatter(0, concentration_b, color="black", s=16)

    concentration_combo_theor = get_combo_additive_expectation(
        e_experimental,
        model_a,
        model_b,
        model_combo,
        model_combo.cocktail.ratio,
        plot=False,
    )
    concentration_combo_theor_a = (
        concentration_combo_theor * model_combo.cocktail.ratio.to_proportion()
    )
    concentration_combo_theor_b = (
        concentration_combo_theor
        * model_combo.cocktail.ratio.reciprocal().to_proportion()
    )

    plt.scatter(
        concentration_combo_theor_a,
        concentration_combo_theor_b,
        color="lightslategrey",
        label=f"Expected Equipotent Combo, $EC_{{{(e_experimental * 100):.0f}}}$",
        s=16,
    )
    print(
        (
            f"Theoretical equipotent combo: "
            f"({concentration_combo_theor_a}{model_a.get_x_units()} {model_a.cocktail}, "
            f"{concentration_combo_theor_b}{model_b.get_x_units()} {model_b.cocktail})"
        )
    )

    concentration_combo_exper = model_combo.effective_concentration(e_experimental)
    concentration_combo_exper_a = (
        concentration_combo_exper * model_combo.cocktail.ratio.to_proportion()
    )
    concentration_combo_exper_b = (
        concentration_combo_exper
        * model_combo.cocktail.ratio.reciprocal().to_proportion()
    )

    color = (
        "tab:red"
        if concentration_combo_exper > concentration_combo_theor
        else "tab:green"
    )
    plt.scatter(
        concentration_combo_exper_a,
        concentration_combo_exper_b,
        color=color,
        label=f"Observed Equipotent Combo, $EC_{{{(e_experimental * 100):.0f}}}$",
        s=16,
    )
    print(
        (
            f"Observed equipotent combo: "
            f"({concentration_combo_exper_a}{model_a.get_x_units()} {model_a.cocktail}, "
            f"{concentration_combo_exper_b}{model_b.get_x_units()} {model_b.cocktail})"
        )
    )

    fic = get_combo_FIC(
        e_experimental, model_a, model_b, model_combo, model_combo.cocktail.ratio
    )
    print(f"FIC_{(e_experimental * 100):.0f}={fic:.2f} for {model_combo.cocktail}")
    offset_x = (
        max(model_a.xs) / 64
    )  # arbitrary adjustments to put text in nice location
    offset_y = max(model_b.xs) / 128
    ax.annotate(
        f"$FIC_{{{(e_experimental * 100):.0f}}}={fic:.2f}$",
        xy=(
            concentration_combo_exper_a + offset_x,
            concentration_combo_exper_b - offset_y,
        ),
        textcoords="data",
    )

    inhibition_max_a = 1 - model_a.get_pct_survival(ys=model_a.c)
    inhibition_max_b = 1 - model_a.get_pct_survival(ys=model_b.c)

    def f_isobole(ec_combo_a):
        return do_additive_isobole(
            ec_combo_a,
            model_a.e,
            model_b.e,
            inhibition_max_a,
            inhibition_max_b,
            concentration_b,
            model_b.b,
            model_a.b,
        )

    plot_func(
        model_a.xs,
        f_isobole,
        f"{model_combo.cocktail} $EC_{{{(e_experimental * 100):.0f}}}$",
        None,
        close=False,
        color="tab:gray",
        max_x=concentration_a,
        max_y=concentration_b,
        min_x=0,
        min_y=0,
    )

    max_x = max(
        concentration_a + offset_x,
        concentration_combo_theor_a + offset_x,
        concentration_combo_exper_a + offset_x,
        max_x,
    )
    max_y = max(
        concentration_b + offset_y,
        concentration_combo_theor_b + offset_y,
        concentration_combo_exper_b + offset_y,
        max_y,
    )

    uniq_str = str(int(time() * 1000) % 1_620_000_000_000)
    log_dir = f'{get_config_setting("log_dir")}/dose_response'
    os.makedirs(log_dir, exist_ok=True)
    return (
        os.path.join(
            f'{get_config_setting("log_dir")}/dose_response',
            f"{model_a.cocktail}+{model_b.cocktail}_isoboles_{uniq_str}.png",
        ),
        max_x,
        max_y,
    )


def chart_checkerboard(model_a, model_b, models_combo, file_name_context=None):
    file_name_context1 = ""
    file_name_context2 = ""
    if file_name_context:
        file_name_context1 = "(" + file_name_context + ")"
        file_name_context2 = file_name_context + "_"

    label_a = f"{model_a.cocktail} Concentration ({model_a.get_x_units()})"
    label_b = f"{model_b.cocktail} Concentration ({model_b.get_x_units()})"

    data_dict = {}
    data_dict[label_a] = [float(x) for x in model_a.xs]
    data_dict[label_b] = [0] * len(model_a.xs)
    data_dict["Pct. Survival"] = [
        float(y) for y in model_a.get_pct_survival(ys=model_a.ys)
    ]
    data_dict[label_a] = np.append(data_dict[label_a], [0] * len(model_b.xs))
    data_dict[label_b] = np.append(data_dict[label_b], [float(x) for x in model_b.xs])
    data_dict["Pct. Survival"] = np.append(
        data_dict["Pct. Survival"], model_b.get_pct_survival(ys=model_b.ys)
    )

    for model_combo in models_combo:
        xs = np.array([float(x) for x in model_combo.xs])
        data_dict[label_a] = np.append(
            data_dict[label_a], xs * model_combo.cocktail.ratio.to_proportion()
        )
        data_dict[label_b] = np.append(
            data_dict[label_b],
            xs * model_combo.cocktail.ratio.reciprocal().to_proportion(),
        )
        data_dict["Pct. Survival"] = np.append(
            data_dict["Pct. Survival"], model_combo.get_pct_survival(ys=model_combo.ys)
        )

    data = pd.DataFrame(data_dict)
    data = data.pivot_table(
        index=label_a, columns=label_b, values="Pct. Survival", aggfunc="median"
    )

    _fig = plt.figure(figsize=(12, 8), dpi=100)
    ax = sns.heatmap(
        data,
        vmin=0,
        vmax=1,
        cmap="mako",
        annot=True,
        fmt=".0%",
        linewidths=2,
        square=True,
        cbar_kws={
            "format": PercentFormatter(xmax=1, decimals=0),
            "label": "Remaining Hair-Cell Brightness",
            "ticks": [0, 1],
            # 'location': 'bottom',
            # 'shrink': 0.5,
        },
    )
    ax.invert_yaxis()
    plt.title(
        f"{model_a.get_condition()} vs. {model_b.get_condition()} {file_name_context1}"
    )
    plt.tight_layout()
    uniq_str = str(int(time() * 1000) % 1_620_000_000_000)
    log_dir = f'{get_config_setting("log_dir")}/dose_response'
    os.makedirs(log_dir, exist_ok=True)
    plt.savefig(
        f"{log_dir}/{model_a.get_condition()}-{model_b.get_condition()}_{file_name_context2}"
        + f"checkerboard_{uniq_str}.png"
    )
    plt.clf()


def chart_diamond(model_a, model_b, model_combo):
    if model_a.c is None or model_b.c is None:
        print("Cannot chart diamond due to failed dose-response curve fit")
        return

    # chart A and B on the same axes, with the same x values

    model_b_scaled = Model(
        np.array(model_b.xs) * model_combo.cocktail.ratio,
        model_b.ys,
        model_b.cocktail,
        model_b.E_0,
        model_b.E_max,
    )

    model_a.chart(close=False)
    model_b_scaled.chart(
        color="tab:blue",
        label=False,
        name=f"{model_a.cocktail}_w_adj_{model_b.cocktail}_overlay_@{model_combo.cocktail.effect}",
    )

    # heatmap
    data = pd.DataFrame(
        {
            "concentration": list(model_combo.xs) * 3,
            "score": np.concatenate(
                (
                    model_a.get_ys(model_a.xs),
                    model_b.get_ys(model_b.xs),
                    model_combo.get_ys(model_combo.xs),
                )
            ),
            "condition": [model_a.get_condition()] * len(model_a.xs)
            + [model_b.get_condition()] * len(model_b.xs)
            + [model_combo.get_condition()] * len(model_combo.xs),
        }
    )
    data = data.pivot_table(
        index="condition", columns="concentration", values="score", aggfunc="median"
    )

    sns.heatmap(
        data,
        vmin=model_a.get_absolute_E_max(),
        vmax=model_a.E_0,
        cmap="viridis",
        annot=True,
        fmt=".1f",
        linewidths=1,
        square=True,
    )
    plt.title(f"{model_a.get_condition()} vs. {model_b.get_condition()}: Model")
    plt.tight_layout()
    uniq_str = str(int(time() * 1000) % 1_620_000_000_000)
    log_dir = f'{get_config_setting("log_dir")}/dose_response'
    os.makedirs(log_dir, exist_ok=True)
    plt.savefig(
        f"{log_dir}/combo_{model_a.get_condition()}-{model_b.get_condition()}_model_{uniq_str}.png"
    )
    plt.clf()


# derived from Grabovsky and Tallarida 2004, http://doi.org/10.1124/jpet.104.067264, Eq. 3
# where B is the drug with the higher maximum effect = lower survival at maximum effect
# a_i = amount of A in combination required to reach relevant effect level
# b_i = amount of B in combination required to reach relevant effect level
# A_E50_a = amount of A alone required to reach 50% of A's maximum effect
# B_E50_b = amount of B alone required to reach 50% of B's maximum effect
# E_max_a = A's maximum effect
# E_max_b = B's maximum effect
# B_i = amount of B alone required to reach relevant effect level
# p = Hill function coefficient for B's dose-response curve
# q = Hill function coefficient for A's dose-response curve
# returns b_i
def do_additive_isobole(a_i, A_E50_a, B_E50_b, E_max_a, E_max_b, B_i, p, q):
    return B_i - B_E50_b / ((E_max_b / E_max_a) * (1 + A_E50_a**q / a_i**q) - 1) ** (
        1 / p
    )


# derived from Grabovsky and Tallarida 2004, http://doi.org/10.1124/jpet.104.067264, Eq. 3
# for details see above
# returns the FIC score
def do_FIC(a_i, b_i, A_E50_a, B_E50_b, E_max_a, E_max_b, B_i, p, q):
    return (
        b_i + B_E50_b / ((E_max_b / E_max_a) * (1 + A_E50_a**q / a_i**q) - 1) ** (1 / p)
    ) / B_i


def filter_valid(array, minimum=None, tolerance=None):
    if minimum is not None:
        array = [element for element in array if element >= minimum]

    if tolerance is not None:
        blacklist = []
        for i, element in enumerate(array):
            if i not in blacklist:
                for j, element_j in enumerate(array):
                    if (
                        i != j
                        and j not in blacklist
                        and utils.equalsish(element, element_j, delta=tolerance)
                    ):
                        blacklist.append(j)
        blacklist.sort(reverse=True)
        blacklist = dict.fromkeys(blacklist)
        for removable_idx in blacklist:
            del array[removable_idx]

    return array


def get_bliss_ixn(x, y, model_a, model_b, model_combo):
    if float(x) == 0:
        return np.nan

    # concentration_a = x * model_combo.cocktail.ratio.to_proportion()
    # concentration_b = x * model_combo.cocktail.ratio.reciprocal().to_proportion()
    x_doses = {dose.drug: dose for dose in x.doses}
    dose_a = x_doses[model_a.cocktail.drugs[0]]
    dose_b = x_doses[model_b.cocktail.drugs[0]]

    model_a_doses = [solution.doses[0] for solution in model_a.xs]
    model_b_doses = [solution.doses[0] for solution in model_b.xs]

    if dose_a in model_a_doses:
        pct_survival = (
            model_a.ys[model_a_doses.index(dose_a)] - model_a.get_absolute_E_max()
        ) / (model_a.E_0 - model_a.get_absolute_E_max())
        e_inhibition_a = 1 - pct_survival
    else:
        print(f"WARNING: {dose_a} not found in {model_a}. Using modeled response.")
        e_inhibition_a = 1 - model_a.get_pct_survival(xs=float(dose_a))

    if dose_b in model_b_doses:
        pct_survival = (
            model_b.ys[model_b_doses.index(dose_b)] - model_a.get_absolute_E_max()
        ) / (model_b.E_0 - model_b.get_absolute_E_max())
        e_inhibition_b = 1 - pct_survival
    else:
        print(f"WARNING: {dose_b} not found in {model_b}. Using modeled response.")
        e_inhibition_b = 1 - model_b.get_pct_survival(xs=float(dose_b))

    fract_inhib_theor = (
        e_inhibition_a + e_inhibition_b - e_inhibition_a * e_inhibition_b
    )
    fract_inhib_observed = 1 - model_combo.get_pct_survival(ys=y)

    return fract_inhib_observed - fract_inhib_theor


def get_combo_additive_expectation(
    pct_inhibition, model_a, model_b, model_combo, combo_ratio_a, plot=True
):
    # set model_b to the model with the higher maximum effect = lower survival at maximum effect
    if model_a.c < model_b.c:
        model_a, model_b = model_b, model_a
        combo_ratio_a = combo_ratio_a.reciprocal()

    ec_a_alone = model_a.effective_concentration(pct_inhibition)
    ec_b_alone = model_b.effective_concentration(pct_inhibition)

    if np.isnan(ec_b_alone):
        return np.nan

    inhibition_max_a = 1 - model_a.get_pct_survival(ys=model_a.c)
    inhibition_max_b = 1 - model_a.get_pct_survival(ys=model_b.c)

    # find intersection between dose ratio and additive isobole
    def f_isobole(ec_combo_a):
        return do_additive_isobole(
            ec_combo_a,
            model_a.e,
            model_b.e,
            inhibition_max_a,
            inhibition_max_b,
            ec_b_alone,
            model_b.b,
            model_a.b,
        )

    def f_diagonal(ec_combo_a):
        return ec_combo_a / combo_ratio_a

    simplistic_additive_estimate = 1 if np.isnan(ec_a_alone) else ec_a_alone / 2

    if plot:
        plot_func(
            model_a.xs,
            f_isobole,
            f"{model_combo.cocktail} Additive Isobole",
            f"{model_combo.cocktail}_isobole",
            f"{model_a.cocktail} ({model_a.get_x_units()})",
            close=False,
            color="tab:blue",
            max_x=(ec_a_alone if not np.isnan(ec_a_alone) else None),
            min_y=0,
        )
        plot_func(
            model_a.xs,
            f_diagonal,
            f"{model_combo.cocktail} Ratio",
            None,
            close=False,
            color="tab:green",
            max_x=(ec_a_alone if not np.isnan(ec_a_alone) else None),
        )
        plot_func(
            model_a.xs,
            lambda xs: ec_b_alone * (1 - xs / ec_a_alone),
            "Line of simplistic additivity",
            f"{model_combo.cocktail}_isobole",
            color="lightgrey",
            linestyle="dashed",
            min_y=0,
        )

    conc_a = get_intersection(f_isobole, f_diagonal, simplistic_additive_estimate)[0]
    conc_b = conc_a / combo_ratio_a
    return conc_a + conc_b


def get_combo_FIC(
    pct_inhibition, model_a, model_b, model_combo, combo_ratio_a, silent=False
):
    # set model_b to the model with the higher maximum effect = lower survival at maximum effect
    if model_a.c < model_b.c:
        model_a, model_b = model_b, model_a
        combo_ratio_a = combo_ratio_a.reciprocal()

    ec_b_alone = model_b.effective_concentration(pct_inhibition, silent)
    ec_combo = model_combo.effective_concentration(pct_inhibition, silent)

    if np.isnan(ec_b_alone) or np.isnan(ec_combo):
        return np.nan

    ec_combo_a = ec_combo * combo_ratio_a.to_proportion()
    ec_combo_b = ec_combo * combo_ratio_a.reciprocal().to_proportion()
    inhibition_max_a = 1 - model_a.get_pct_survival(ys=model_a.c)
    inhibition_max_b = 1 - model_a.get_pct_survival(ys=model_b.c)

    return do_FIC(
        ec_combo_a,
        ec_combo_b,
        model_a.e,
        model_b.e,
        inhibition_max_a,
        inhibition_max_b,
        ec_b_alone,
        model_b.b,
        model_a.b,
    )


def get_intersection(f1, f2, guess):
    def f_intersection_equals_zero(xs):
        return np.array(f1(xs), dtype=np.float64) - np.array(f2(xs), dtype=np.float64)

    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            return scipy.optimize.root(f_intersection_equals_zero, guess, method="lm").x
    except scipy.optimize.nonlin.NoConvergence as e:
        return e.args[0]


# Ritz 2009, https://doi.org/10.1002/etc.7, Eq. 2
# `xs` is a numpy array of x values; b, c, d, and e are model parameters:
# relative slope at inflection point, lower asymptote, upper asymptote, inflection point (EC_50)
# returns y values
def log_logistic_model(xs, b, c, d, e):
    return c + (d - c) / (1 + (xs / e) ** b)


def neo_E_max():
    _neo_model_local = _get_neo_model()
    return _neo_model_local.get_condition_E_max()


def plot_func(
    xs,
    func,
    label,
    filename_prefix,
    x_label=None,
    close=True,
    color="darkgrey",
    linestyle="solid",
    max_x=None,
    max_y=None,
    min_x=None,
    min_y=None,
    y_label=None,
):
    if x_label:
        plt.xlabel(x_label)
    if y_label:
        plt.ylabel(y_label)
    max_x = float(max_x) if max_x is not None else float(max(xs))
    min_x = float(min_x) if min_x is not None else float(min(xs))
    line_xs = np.linspace(min_x, max_x, 100)
    if max_x is not None:
        line_xs = line_xs[line_xs <= float(max_x)]
    if min_x is not None:
        line_xs = line_xs[line_xs >= float(min_x)]
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        line_ys = func(line_xs)
    if max_y is not None:
        line_xs = line_xs[line_ys <= max_y]
        line_ys = line_ys[line_ys <= max_y]
    if min_y is not None:
        line_xs = line_xs[line_ys >= min_y]
        line_ys = line_ys[line_ys >= min_y]
    plt.plot(line_xs, line_ys, color=color, label=label, marker=None, zorder=-1)
    plt.legend()
    if close:
        plt.tight_layout()
        uniq_str = str(int(time() * 1000) % 1_620_000_000_000)
        log_dir = f'{get_config_setting("log_dir")}/dose_response'
        os.makedirs(log_dir, exist_ok=True)
        plt.savefig(os.path.join(log_dir, f"{filename_prefix}_{uniq_str}.png"))
        plt.close()
        plt.clf()


def _get_model(filename, debug=1, opened: bool = False):
    xs, ys = [], []

    if not opened:
        with open(filename, encoding="utf8", newline="") as f:
            for x, y in csv.reader(f, delimiter="\t"):
                xs.append(utils.Solution(x))
                ys.append(float(y))
    else:
        for x, y in csv.reader(filename, delimiter="\t"):
            xs.append(utils.Solution(x))
            ys.append(float(y))

    model = Model(xs, ys, xs[-1].get_cocktail(), E_max=np.nanmin(ys))
    if debug:
        model.chart()
    return model


def _get_neo_model(debug=1):
    global _neo_model
    if _neo_model is None:
        with importlib.resources.open_text(
            "pepitatools.data", "neo_data.csv"
        ) as data_file:
            _neo_model = _get_model(data_file, debug, opened=True)
        # _neo_model = _get_model(os.path.join(utils.get_here(), 'examples/neo_data.csv'), debug)
    return _neo_model



# Imports
# Standard Library Imports
import copy
import json
import math
import os
from pathlib import Path
import re
from time import time
import warnings

# External Imports
from matplotlib.ticker import PercentFormatter
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

# Local Imports
from . import absolute, analyze, dose_response, interactions, utils
from .configuration import get_config_setting

ALPHA = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
NUMS = [str(n) for n in range(1, 99)]


def adjust_absolute_filename(filename):
    filepath = Path(filename)
    return filepath.parent.joinpath(filepath.stem + "_absolute" + filepath.suffix)


def generate_plate_schematic(
    schematic,
    results,
    conversions=None,
    plate_info="[Unknown]",
    scale=None,
    well_count=96,
    cmap="mako",
    max_val=100,
):
    results = copy.deepcopy(results)  # don't mess with `results`, it's used after this

    # scale values, and adjust heatmap settings accordingly, if necessary

    if scale is not None:
        min_, max_ = scale
        for solution, values in results.items():
            for i in range(len(values)):
                values[i] = (values[i] - min_) / (max_ - min_)

    vmax, hmap_fmt, cbar_fmt, cbar_label = (
        (
            1,
            ".0%",
            PercentFormatter(xmax=1, decimals=0),
            "Remaining Hair-Cell Brightness",
        )
        if scale is not None
        else (max_val, ".0f", None, "Pipeline Score")
    )

    # produce matrix of responses by combining data from `schematic` and `results`

    # iterate backwards so as not to mess up indices for subsequent loops
    for row_idx in reversed(range(len(schematic))):
        if not schematic[row_idx]:
            del schematic[
                row_idx
            ]  # remove empty lists or numpy will reject ragged array

    height = len(schematic)
    max_width = max([len(schematic[row_idx]) for row_idx in range(height)])

    annotations = np.full((height, max_width), "", dtype=object)
    responses = np.full((height, max_width), np.nan, dtype=np.double)

    for row_idx in range(height):
        for col_idx in range(len(schematic[row_idx])):
            solution = utils.Solution(schematic[row_idx][col_idx], conversions)
            result = results[solution].pop(0)
            label = re.sub(
                r"([A-Z])([A-Za-z])\w+\s?([\d./]+)?([A-Za-zμ ]+)?",
                lambda m: f'{m.group(1)}{m.group(2).lower()}{m.group(3) if m.group(3) else ""}',
                schematic[row_idx][col_idx],
            )
            label = re.sub(r"\s+", "", label)
            label = re.sub(r"\+", "+\n", label)
            annotations[row_idx, col_idx] = f"{result:{hmap_fmt}}\n{label}"
            responses[row_idx, col_idx] = result

    plate_height = utils.plate_height(well_count)
    plate_width = well_count // plate_height

    col_labels = NUMS[1 : plate_width - 1]
    row_labels = list(ALPHA[1 : plate_height - 1])

    plates_count_est = math.ceil(len(responses) / plate_height)

    row_labels *= plates_count_est

    for i in range(
        plates_count_est - 1, 0, -1
    ):  # insert blank line(s) for any plates > 1
        insertable_idx = (len(responses) // plates_count_est) * i
        annotations = np.insert(
            annotations, insertable_idx, np.full_like(annotations[0], ""), axis=0
        )
        responses = np.insert(
            responses, insertable_idx, np.full_like(responses[0], np.nan), axis=0
        )
        row_labels = row_labels[0:insertable_idx] + [""] + row_labels[insertable_idx:]

    # make the heatmap

    _fig = plt.figure(figsize=(12, 6 * plates_count_est), dpi=100)

    ax = sns.heatmap(
        responses,
        vmin=0,
        vmax=vmax,
        cmap=cmap,
        annot=annotations,
        fmt="",
        linewidths=2,
        square=True,
        cbar_kws={
            "format": cbar_fmt,
            "label": cbar_label,
            "ticks": [0, vmax],
        },
        xticklabels=col_labels,
        yticklabels=row_labels,
    )
    ax.tick_params(top=True, bottom=False, labeltop=True, labelbottom=False)

    if max_val < 100:
        suffix1, suffix2 = ", Log", "_log"
    elif scale is None:
        suffix1, suffix2 = "", ""
    elif scale[1] > 10_000:
        suffix1, suffix2 = ", Absolute", "_absolute"
    else:
        suffix1, suffix2 = ", Scaled", "_scaled"

    plt.title(f"{plate_info} {well_count}-well Plate Schematic{suffix1}")
    uniq_str = str(int(time() * 1000) % 1_620_000_000_000)
    log_dir = f'{get_config_setting("log_dir")}/dose_response'
    plt.savefig(
        f"{log_dir}/{plate_info}_{well_count}-well_schematic_heatmap{suffix2}_{uniq_str}.png"
    )
    plt.clf()


def main(
    imagefiles,
    cap=-1,
    chartfile=None,
    checkerboard=False,
    conversions=None,
    debug=0,
    group_regex=".*",
    platefile=None,
    plate_control=None,
    plate_ignore=None,
    plate_info=None,
    plate_positive_control=None,
    treatment_platefile=None,
    absolute_chart=False,
    silent=False,
    talk=False,
):
    if conversions is None:
        conversions = []
    if plate_control is None:
        plate_control = ["B"]
    if plate_ignore is None:
        plate_ignore = []
    if plate_positive_control is None:
        plate_positive_control = []
    hashfile = utils.get_inputs_hashfile(
        imagefiles=imagefiles,
        cap=cap,
        group_regex=group_regex,
        platefile=platefile,
        plate_control=plate_control,
        plate_ignore=plate_ignore,
    )

    if talk:
        sns.set_context("talk")

    conversions = dict(conversions)
    schematic = analyze.get_schematic(
        platefile, len(imagefiles), plate_ignore, flat=False
    )

    if absolute_chart:
        abs_chartfile = adjust_absolute_filename(chartfile)
        results2 = absolute.main(
            imagefiles,
            cap=cap,
            chartfile=abs_chartfile,
            debug=0,
            group_regex=group_regex,
            platefile=platefile,
            plate_control=plate_control,
            plate_ignore=plate_ignore,
            silent=False,
        )
        results2 = {
            utils.Solution(key, conversions): value for key, value in results2.items()
        }
        abs_max = int(get_config_setting("absolute_max_ototox"))
        abs_min = int(get_config_setting("absolute_min_ototox"))
        generate_plate_schematic(
            schematic,
            results2,
            conversions=conversions,
            plate_info=plate_info,
            scale=(abs_min, abs_max),
            well_count=96,
        )

    if chartfile is None and debug == 0 and os.path.exists(hashfile):
        with open(hashfile, "r") as f:  # read cached results
            results = json.load(f)
    else:
        results = analyze.main(
            imagefiles,
            cap,
            chartfile,
            debug,
            group_regex,
            platefile,
            plate_control,
            plate_ignore,
            silent=False,
        )
        with open(hashfile, "w") as f:  # cache results for reuse
            json.dump(results, f)

    drug_conditions = _parse_results(results, conversions)
    control_drugs = [
        utils.Cocktail(utils.Dose(control).drug) for control in plate_control
    ]
    models = {}

    results = {
        utils.Solution(key, conversions): value for key, value in results.items()
    }

    # positive control

    positive_control_solutions = [
        utils.Solution(positive_control, conversions)
        for positive_control in plate_positive_control
    ]
    positive_control_scores = [
        result
        for solution in positive_control_solutions
        for result in results[solution]
    ]
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        positive_control_value = np.nanmean(positive_control_scores)

    if np.isnan(positive_control_value):
        print(
            (
                "WARNING: No positive control included. Using minimum calculated value as "
                "positive control"
            )
        )
        positive_control_value = np.nanmin(
            [value for condition, values in results.items() for value in values]
        )

    # generate plate schematics

    generate_plate_schematic(
        schematic,
        results,
        conversions=conversions,
        plate_info=plate_info,
        scale=(positive_control_value, 100),
        well_count=96,
    )

    # generate models, dose-response charts

    for cocktail, conditions in drug_conditions.items():
        if cocktail in control_drugs:
            continue
        cocktail_scores = {}
        summary_scores = []
        for control_drug in control_drugs:
            for solution in drug_conditions[control_drug]:
                conditions.insert(0, solution)
        for solution in conditions:
            cocktail_scores[solution] = results[solution]
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", RuntimeWarning)
                summary_score = np.nanmedian(results[solution])
                summary_scores.append(summary_score)
        models[cocktail] = dose_response.Model(
            conditions, summary_scores, cocktail, E_max=positive_control_value
        )
        models[cocktail].chart(
            results[solution],
            datapoints=cocktail_scores,
            name=plate_info + "_" + str(cocktail) if plate_info else None,
            scale=[positive_control_value, 100],
        )

    # print EC values

    for model in models.values():
        for ec_value in (50, 75, 90, 99):
            concentn = model.effective_concentration(ec_value / 100)
            if not np.isnan(concentn):
                print(
                    (
                        f"{model.get_condition()} "
                        f"EC_{ec_value}={concentn:.2f}{model.get_x_units()}"
                    )
                )

    # analyze combinations

    models_combo = [model for model in models.values() if model.combo]

    if not checkerboard:
        total_max_x = 1
        total_max_y = 1

        fig = plt.figure(figsize=(12, 8), dpi=100)
        ax = fig.add_subplot(1, 1, 1)
        ax.margins(0.006)

        for model_combo in models_combo:
            subcocktail_a = utils.Cocktail(model_combo.cocktail.drugs[0])
            if subcocktail_a not in models:
                continue
            subcocktail_b = utils.Cocktail(model_combo.cocktail.drugs[1])
            model_a = models[subcocktail_a]
            model_b = models[subcocktail_b]
            plot_filename, max_x, max_y = dose_response.analyze_diamond(
                model_a, model_b, model_combo
            )
            dose_response.chart_diamond(model_a, model_b, model_combo)

            total_max_x = max(total_max_x, max_x)
            total_max_y = max(total_max_y, max_y)

        if models_combo and plot_filename is not None:
            plt.xlim(right=total_max_x)
            plt.ylim(top=total_max_y)
            plt.savefig(plot_filename)
            plt.close()
            plt.clf()
    else:
        pairs = {
            (model_combo.cocktail.drugs[0], model_combo.cocktail.drugs[1])
            for model_combo in models_combo
        }

        for pair in pairs:
            model_a, model_b = (
                models[utils.Cocktail(pair[0])],
                models[utils.Cocktail(pair[1])],
            )

            models_combo_relevant = [
                model_combo
                for model_combo in models_combo
                if model_combo.cocktail.drugs[0] in pair
                and model_combo.cocktail.drugs[1] in pair
            ]

            dose_response.analyze_checkerboard(
                model_a,
                model_b,
                models_combo_relevant,
                method="Bliss",
                file_name_context=plate_info,
            )
            dose_response.chart_checkerboard(
                model_a, model_b, models_combo_relevant, file_name_context=plate_info
            )

            doses_a = np.array(
                [x.doses[0] for x in model_a.xs if x.get_drugs() != ("Control",)]
            )
            doses_b = np.array(
                [x.doses[0] for x in model_b.xs if x.get_drugs() != ("Control",)]
            )

            responses_all_a = squarify(
                [results[x] for x in model_a.xs if x.get_drugs() != ("Control",)]
            )
            responses_all_b = squarify(
                [results[x] for x in model_b.xs if x.get_drugs() != ("Control",)]
            )

            combo_solutions = [
                solution for solution in results.keys() if len(solution.doses) > 1
            ]

            doses_a_ab = np.array([solution.doses[0] for solution in combo_solutions])
            doses_b_ab = np.array([solution.doses[1] for solution in combo_solutions])
            responses_all_ab = squarify(
                [results[solution] for solution in combo_solutions]
            )

            if not positive_control_scores:
                positive_control_scores = np.array(
                    [
                        min(
                            [
                                result
                                for results_list in results.values()
                                for result in results_list
                            ]
                        )
                    ]
                )

            if not plate_info:
                plate_info = os.path.basename(
                    os.path.dirname(os.path.dirname(imagefiles[0]))
                )

            try:
                interactions.response_surface(
                    doses_a,
                    responses_all_a,
                    doses_b,
                    responses_all_b,
                    doses_a_ab,
                    doses_b_ab,
                    responses_all_ab,
                    positive_control_scores,
                    sampling_iterations=1000,
                    sample_size=20,
                    model_size=1,
                    alpha=0.1,
                    file_name_context=plate_info,
                )
            except ValueError as ve:
                if ve.args[0] == "cov must be 2 dimensional and square":
                    print("Unable to analyze non-square interaction matrix")
                else:
                    raise ve


def squarify(list_of_lists):
    width = max([len(row) for row in list_of_lists])

    for row in list_of_lists:
        pad_size = width - len(row)
        row.extend([np.nan for _ in range(pad_size)])

    return np.array(list_of_lists)


def _key_value_pair(argument, delimiter="="):
    return tuple(argument.split(delimiter))


def _parse_results(results, conversions):
    drug_conditions = {}
    for condition in results:
        solution = utils.Solution(condition, conversions)
        utils.put_multimap(drug_conditions, solution.get_cocktail(), solution)
    return drug_conditions



"""
Utility functions for PEPITA-tools
"""

# Imports
# Standard Library Imports
from __future__ import annotations
import base64
import hashlib
import math
import os
import pickle
import re

# External Imports
import numpy as np

# Local Imports
from .configuration import get_config_setting


class Cocktail:
    def __eq__(self, other):
        return self.drugs == other.drugs and self.ratio == other.ratio

    def __hash__(self):
        return hash(self.drugs) ^ hash(self.ratio)

    def __init__(self, drugs, effect=None, ratio=None):
        self.drugs = drugs if not isinstance(drugs, str) else (drugs,)
        self.effect = effect
        self.ratio = ratio

    def __repr__(self):
        string = "+".join(self.drugs)
        if self.ratio:
            string += f"@{self.ratio}"
            if (
                self.effect
            ):  # effect level only relevant if ratio is set, otherwise confusing
                string += f"(EC{self.effect})"
        return string


class Dose:
    _ec_pattern = re.compile(r"([0-9]*)([A-Z]{3}[0-9]{2})/?([0-9]*)")
    _vector_pattern = re.compile(r"(.+?) ([0-9]+[.]?[0-9]*) ?([^0-9]+)")

    def __add__(self, other):
        return Dose(f"{self.drug} {self.quantity + float(other)}{self.unit}")

    def __eq__(self, other):
        if not isinstance(other, Dose):
            return False
        return (
            self.drug == other.drug
            and self.quantity == other.quantity
            and self.unit == other.unit
        )

    def __float__(self):
        return self.quantity

    def __hash__(self):
        return hash(self.drug) ^ hash(self.quantity) ^ hash(self.unit)

    def __init__(self, string, conversions=None):
        if conversions is None:
            conversions = {}

        self.converted = False
        self.ec = False
        self.string = string

        ec_match = Dose._ec_pattern.fullmatch(string)

        if ec_match:
            self.ec = True
            multiplier, ec_data, divisor = ec_match.group(1, 2, 3)
            self.series = ec_data
            if ec_data in conversions:
                self.converted = True
                ec_data = conversions[ec_data]
            vector_match = Dose._vector_pattern.fullmatch(ec_data)
        elif string in conversions:
            vector_match = Dose._vector_pattern.fullmatch(conversions[string])
            self.converted = True
            self.series = None
            multiplier, divisor = None, None
        else:
            vector_match = Dose._vector_pattern.fullmatch(string)
            self.series = None
            multiplier, divisor = None, None

        if vector_match:
            self.drug, self.quantity, self.unit = vector_match.group(1, 2, 3)
            self.quantity = float(self.quantity)
            if not self.series:
                self.series = vector_match.group(1)
        else:
            raise ValueError(
                f'Dose string "{string}" unknown or not in the proper format'
            )

        if multiplier:
            self.quantity *= int(multiplier)
        if divisor:
            self.quantity /= int(divisor)

    def __mul__(self, other):
        return Dose(f"{self.drug} {self.quantity * other}{self.unit}")

    def __radd__(self, other):
        return Dose(f"{self.drug} {float(other) + self.quantity}{self.unit}")

    def __repr__(self):
        return f"{self.drug} {self.quantity}{self.unit}"


class Ratio:
    def __eq__(self, other):
        if isinstance(other, Ratio):
            return (self.num * other.denom) == (other.num * self.denom)
        else:
            return float(self) == float(other)

    def __float__(self):
        return self.num / self.denom

    def __hash__(self):
        return hash(round(self.num / self.denom, 8))

    def __init__(self, num, denom):
        self.num = num
        self.denom = denom

    def __mul__(self, other):
        return round(other * self.num / self.denom, 8)

    def __rmul__(self, other):
        return round(other * self.num / self.denom, 8)

    def __repr__(self):
        return f"{self.num}:{self.denom}"

    def __rtruediv__(self, other):
        return other * self.reciprocal()

    def reciprocal(self):
        return Ratio(self.denom, self.num)

    def to_proportion(self):
        return Ratio(self.num, self.num + self.denom)


class Solution:
    def __init__(self, string, conversions=None):
        if conversions is None:
            conversions = []

        self.conversions = conversions
        self.string = string
        dose_strings = [s.strip() for s in string.split("+")]
        self.doses = [Dose(string, conversions) for string in dose_strings]

    def __eq__(self, other):
        if not isinstance(other, Solution):
            return False
        return self.doses == other.doses

    def __float__(self):
        return float(sum(self.doses))

    def __gt__(self, other):
        return float(self) > float(other)

    def __hash__(self):
        return hash(tuple(self.doses))

    def __mul__(self, other):
        doses = [dose * other for dose in self.doses]
        return Solution(" + ".join([dose.string for dose in doses]))

    def __repr__(self):
        return "+".join(
            f"{dose.drug} {dose.quantity}{dose.unit}" for dose in self.doses
        )

    def __rmul__(self, other):
        return other * float(self)

    def __truediv__(self, other):
        return float(self) / other

    def combine_doses(self, other):
        return Solution(f"{self.string} + {other.string}")

    def dilute(self, dilution):
        if dilution < 0 or dilution > 1:
            raise ValueError("Solution should be diluted by a factor between 0 and 1")
        doses = [dose * dilution for dose in self.doses]
        return Solution(" + ".join([dose.string for dose in doses]))

    def get_cocktail(self):
        effect = extract_number(self.doses[0].series)
        return Cocktail(
            tuple(dose.drug for dose in self.doses),
            effect=(None if np.isnan(effect) else effect),
            ratio=(None if len(self.doses) != 2 else self.ratio()),
        )

    def get_drugs(self):
        return tuple(dose.drug for dose in self.doses)

    def get_units(self):
        return "+".join(str(dose.unit) for dose in self.doses)

    def ratio(self):
        if len(self.doses) == 2:
            return Ratio(self.doses[0].quantity, self.doses[1].quantity)
        raise ValueError(f"This solution {self.doses} does not have a valid dose ratio")

    def reverse(self):
        doses = self.doses[::-1]
        return Solution(" + ".join([str(dose) for dose in doses]), self.conversions)


def equalsish(val1, val2, delta=0.001):
    return abs(val1 - val2) < delta


def extract_number(string):
    number_match = re.search(r"[0-9.]+", string)
    if number_match:
        return float(number_match.group(0))
    else:
        return np.nan


def geometric_mean(array):
    return np.exp(np.mean(np.log(array)))


# def get_config(setting, fallback=None):
# 	global _config
# 	if _config == None:
# 		_config = configparser.ConfigParser()
# 		_config.read(f'{get_here()}/config.ini')
# 		_config.read(f'{get_here()}/config-ext.ini')
# 	return _config[_section].get(setting, fallback)


def get_inputs_hashfile(**kwargs):
    sha1hash = hashlib.sha1()
    for value in kwargs.values():
        sha1hash.update(pickle.dumps(value))
    digest = base64.b32encode(sha1hash.digest()).decode("utf-8")
    os.makedirs(os.path.join(get_config_setting("log_dir"), ".cache"), exist_ok=True)
    return os.path.join(get_config_setting("log_dir"), ".cache", f".{digest}.json")


def plate_height(well_count):
    sqrt = int(math.sqrt(well_count))

    for i in range(sqrt, 1, -1):
        if well_count % i == 0:
            return i

    print(f"WARNING: proper plate height not found for {well_count}-well plate")
    return sqrt


def put_multimap(dict_, key, value):
    list_ = dict_.get(key, [])
    list_.append(value)
    dict_[key] = list_


def remove_argument(parser, arg):
    removed = False

    for action in parser._actions:
        if (
            action.option_strings and arg in action.option_strings
        ) or action.dest == arg:
            parser._remove_action(action)
            removed = True
            break

    for action in parser._action_groups:
        for group_action in action._group_actions:
            if group_action.dest == arg:
                action._group_actions.remove(group_action)
                removed = True
                break

    return removed


def remove_arguments(parser, *args):
    return [remove_argument(parser, arg) for arg in args]



# Imports
# Standard Library Imports
import unittest
import pathlib

# External Imports
import numpy as np

# Local Imports
from pepitatools import utils
from pepitatools.configuration import set_config_setting


class TestUtils(unittest.TestCase):
    def test_equalsish(self):
        self.assertTrue(utils.equalsish(2, 2))
        self.assertTrue(utils.equalsish(1 / 7, 1 / 7))
        self.assertTrue(utils.equalsish(0.000001, 0.000002))

        self.assertFalse(utils.equalsish(0.01, 0.02))
        self.assertFalse(utils.equalsish(1000, -1000))

    def test_extract_number(self):
        self.assertTrue(np.isnan(utils.extract_number("abc")))
        self.assertTrue(utils.extract_number("easy as 123") == 123)
        self.assertTrue(
            utils.extract_number("remember remember the 5th of November") == 5
        )
        self.assertTrue(utils.extract_number("4-score and 20 years ago") == 4)
        self.assertTrue(utils.extract_number("pi is approximately 3.14159") == 3.14159)

    def test_get_inputs_hashfile(self):
        log_dir = pathlib.Path(__file__).parent / "tmp_logs"
        log_dir.mkdir(exist_ok=True)
        set_config_setting("log_dir", str(log_dir))

        self.assertTrue(
            utils.get_inputs_hashfile(dummy1=1, dummy2="two", dummy3=3.0)
            == utils.get_inputs_hashfile(dummy1=1, dummy2="two", dummy3=3.0)
        )
        self.assertTrue(
            utils.get_inputs_hashfile(dummy1=1, dummy2="two", dummy3=3.0)
            != utils.get_inputs_hashfile(dummy1=1, dummy2="two", dummy3=4.0)
        )

        # Delete the temporary log directory, and its contents
        for f in log_dir.glob("*"):
            # Clear out the cache directory
            if f.name == ".cache":
                for sub_f in f.glob("*"):
                    sub_f.unlink()
                f.rmdir()
                continue
            f.unlink()
        log_dir.rmdir()

    def test_put_multimap(self):
        dict_ = {}
        utils.put_multimap(dict_, "key", "value")
        self.assertDictEqual(dict_, {"key": ["value"]})

    def test_cocktail(self):
        self.assertEqual(utils.Cocktail("A"), utils.Cocktail("A"))
        self.assertNotEqual(utils.Cocktail("A"), utils.Cocktail("B"))
        self.assertEqual(
            utils.Cocktail(("A", "B"), 50, utils.Ratio(1, 1)),
            utils.Cocktail(("A", "B"), 50, utils.Ratio(1, 1)),
        )
        self.assertNotEqual(
            utils.Cocktail(("A", "B"), 50, utils.Ratio(1, 1)),
            utils.Cocktail(("A", "C"), 50, utils.Ratio(1, 1)),
        )
        self.assertNotEqual(
            utils.Cocktail(("A", "B"), 50, utils.Ratio(1, 1)),
            utils.Cocktail(("A", "B"), 50, utils.Ratio(2, 1)),
        )

    def test_dose(self):
        assert utils.Dose("XYZ99", conversions={"XYZ99": "XYZ 1μM"}).drug == "XYZ"
        assert utils.Dose("XYZ99", conversions={"XYZ99": "XYZ 1μM"}).quantity == 1
        assert utils.Dose("XYZ99", conversions={"XYZ99": "XYZ 1μM"}).unit == "μM"
        assert utils.Dose("XYZ 1μM").drug == "XYZ"
        assert utils.Dose("XYZ 1μM").quantity == 1
        assert utils.Dose("XYZ 1μM").unit == "μM"
        assert utils.Dose("XYZ99", conversions={"XYZ99": "XYZ 1μM"}) == utils.Dose(
            "XYZ 1μM"
        )
        assert utils.Dose("XYZ99", conversions={"XYZ99": "XYZ 1μM"}) != utils.Dose(
            "XYZ 2μM"
        )
        assert utils.Dose("XYZ 1μg/mL").unit == "μg/mL"

        assert float(utils.Dose("XYZ 1μM")) == 1
        assert float(utils.Dose("XYZ99", conversions={"XYZ99": "XYZ 1μM"})) == 1

        assert utils.Dose("XYZ 1μM") + 1 == utils.Dose("XYZ 2μM")
        assert 1 + utils.Dose("XYZ 1μM") == utils.Dose("XYZ 2μM")
        assert utils.Dose("XYZ 1μM") + utils.Dose("XYZ 1μM") == utils.Dose("XYZ 2μM")

        assert utils.Dose("XYZ 1μM") * 1 == utils.Dose("XYZ 1μM")
        assert utils.Dose("XYZ 1μM") * 2 == utils.Dose("XYZ 2μM")
        assert utils.Dose("XYZ 3μM") * 4 == utils.Dose("XYZ 12μM")
        assert utils.Dose("XYZ 8μM") * 0.5 == utils.Dose("XYZ 4μM")

        assert utils.Dose("XYZ99/2", conversions={"XYZ99": "XYZ 1μM"}).quantity == 0.5
        assert utils.Dose("2XYZ99", conversions={"XYZ99": "XYZ 1μM"}).quantity == 2
        assert utils.Dose("XYZ99/4", conversions={"XYZ99": "XYZ 16μM"}).quantity == 4
        assert utils.Dose("XYZ99/2", conversions={"XYZ99": "XYZ 16μM"}).quantity == 8
        assert utils.Dose("2XYZ99", conversions={"XYZ99": "XYZ 16μM"}).quantity == 32
        assert utils.Dose("3XYZ99", conversions={"XYZ99": "XYZ 16μM"}).quantity == 48

    def test_ratio(self):
        assert utils.Ratio(1, 2) == utils.Ratio(1, 2)
        assert utils.Ratio(1, 2) == utils.Ratio(2, 4)
        assert utils.Ratio(12.5, 5) == utils.Ratio(25, 10)
        assert utils.Ratio(1, 2) != utils.Ratio(1, 4)
        assert utils.Ratio(1, 2) != utils.Ratio(2, 2)
        assert utils.Ratio(3.3, 10) != utils.Ratio(1, 3)
        assert utils.Ratio(1, 2) == 0.5
        assert utils.Ratio(3, 1) == 3

        assert utils.equalsish(float(utils.Ratio(5, 20)), 0.25)
        assert utils.equalsish(float(utils.Ratio(6, 10)), 0.6)
        assert utils.equalsish(float(utils.Ratio(11, 5)), 2.2)

        assert utils.Ratio(1, 4) * utils.Ratio(4, 1) == 1
        assert utils.Ratio(1, 4) * utils.Ratio(4, 1) == utils.Ratio(1, 1)
        assert 3 * utils.Ratio(4, 1) == 12
        assert 2 * utils.Ratio(3, 2) == 3
        assert utils.Ratio(4, 1) * 3 == 12
        assert utils.Ratio(3, 2) * 2 == 3
        assert utils.equalsish(utils.Ratio(12.5, 5) * 5, 12.5)
        assert utils.equalsish(5 * utils.Ratio(12.5, 5), 12.5)

        assert utils.Ratio(1, 2).reciprocal() == utils.Ratio(2, 1)
        assert utils.Ratio(7, 3).reciprocal() == utils.Ratio(3, 7)

        assert 2 / utils.Ratio(1, 2) == 4
        assert 5 / utils.Ratio(5, 2) == 2
        assert 4 / utils.Ratio(8, 1) == 0.5

        assert utils.Ratio(1, 9).to_proportion() == utils.Ratio(1, 10)
        assert utils.Ratio(5, 4).to_proportion() == utils.Ratio(5, 9)

        assert utils.Dose("XYZ 1μM") * utils.Ratio(7, 2) == utils.Dose("XYZ 3.5μM")
        assert utils.Dose("XYZ 6μM") * utils.Ratio(2, 3) == utils.Dose("XYZ 4μM")
        assert utils.Dose("XYZ 32μM") / utils.Ratio(2, 3) == utils.Dose("XYZ 48μM")
        assert utils.Dose("XYZ 12μM") / utils.Ratio(4, 1) == utils.Dose("XYZ 3μM")

    def test_solution(self):
        assert utils.Solution("XYZ 1μM") == utils.Solution("XYZ 1μM")
        assert utils.Solution("ABC 10μg/mL") == utils.Solution("ABC 10μg/mL")
        assert utils.Solution("XYZ 1μM") != utils.Solution("XYZ 1μg/mL")
        assert utils.Solution("XYZ 1μM") != utils.Solution("XYZ 10μM")
        assert utils.Solution("XYZ 1μM") != utils.Solution("ABC 1μM")
        assert utils.Solution(
            "XYZ99", conversions={"XYZ99": "XYZ 1μM"}
        ) == utils.Solution("XYZ 1μM")
        assert utils.Solution("XYZ 1μM + ABC 10μg/mL") == utils.Solution(
            "XYZ 1μM + ABC 10μg/mL"
        )
        assert utils.Solution("XYZ 1μM + ABC 10μg/mL") == utils.Solution(
            "XYZ99 + ABC 10μg/mL", conversions={"XYZ99": "XYZ 1μM"}
        )
        assert utils.Solution("XYZ 1μM").doses[0] == utils.Dose("XYZ 1μM")
        assert utils.Solution("XYZ 1μM + ABC 10μg/mL").doses == [
            utils.Dose("XYZ 1μM"),
            utils.Dose("ABC 10μg/mL"),
        ]

        assert float(utils.Solution("XYZ 1μM + ABC 10μg/mL")) == 11

        assert utils.Solution("XYZ 10μM") > utils.Solution("XYZ 2μM")

        assert utils.Solution("XYZ 1μM") * 2 == utils.Solution("XYZ 2μM")
        assert utils.Solution("XYZ 10μM") * 0.5 == utils.Solution("XYZ 5μM")

        assert 2 * utils.Solution("XYZ 1μM") == 2
        assert 3 * utils.Solution("XYZ 1μM + ABC 10μg/mL") == 33

        assert utils.Solution("XYZ 10μM") / 2 == 5
        assert utils.Solution("XYZ 32μM") / 4 == 8

        assert utils.Solution("XYZ 1μM").combine_doses(
            utils.Solution("ABC 10μg/mL")
        ) == utils.Solution("XYZ 1μM + ABC 10μg/mL")
        assert utils.Solution("XYZ 1μM").combine_doses(
            utils.Solution("ABC 10μg/mL")
        ).doses == [utils.Dose("XYZ 1μM"), utils.Dose("ABC 10μg/mL")]

        assert utils.Solution("XYZ 10μM").dilute(0.5) == utils.Solution("XYZ 5μM")
        assert utils.Solution("XYZ 20μM").dilute(0.2) == utils.Solution("XYZ 4μM")

        assert utils.Solution("XYZ 10μM").get_cocktail() == utils.Cocktail("XYZ")
        assert utils.Solution("XYZ 1μM + ABC 10μg/mL").get_cocktail() == utils.Cocktail(
            ("XYZ", "ABC"), ratio=utils.Ratio(1, 10)
        )


if __name__ == "__main__":
    unittest.main()



# Imports
# Standard Library Imports
import os
from time import time
import warnings

# External Imports
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.optimize
import scipy.stats
import seaborn as sns

# Local Imports
from .configuration import get_config_setting

_rng = np.random.default_rng()


def fit_model_with_noise(
    model_function,
    gamma_guess_0,
    doses_a,
    doses_b,
    doses_a_ab,
    doses_b_ab,
    est_response_covarmat_a,
    est_response_covarmat_b,
    est_true_responses_a,
    est_true_responses_b,
    observed_responses_ab,
):
    noises_a = dict(
        zip(
            doses_a,
            _rng.multivariate_normal(np.zeros(len(doses_a)), est_response_covarmat_a),
        )
    )
    noises_b = dict(
        zip(
            doses_b,
            _rng.multivariate_normal(np.zeros(len(doses_b)), est_response_covarmat_b),
        )
    )

    est_true_responses_a = dict(zip(doses_a, est_true_responses_a))
    est_true_responses_b = dict(zip(doses_b, est_true_responses_b))

    est_theoretical_responses_ab = []

    for dose_a, dose_b in zip(doses_a_ab, doses_b_ab):
        # Zhao 2014, Eq. 5
        est_theoretical_responses_ab.append(
            est_true_responses_a[dose_a]
            + est_true_responses_b[dose_b]
            - est_true_responses_a[dose_a] * est_true_responses_b[dose_b]
            + noises_a[dose_a]
            + noises_b[dose_b]
            - noises_a[dose_a] * noises_b[dose_b]
        )

    doses_a_ab = np.array([float(dose_a) for dose_a in doses_a_ab])
    doses_b_ab = np.array([float(dose_b) for dose_b in doses_b_ab])

    est_theoretical_responses_ab = np.array(est_theoretical_responses_ab)

    return scipy.optimize.least_squares(
        model_function,
        gamma_guess_0,
        args=(
            doses_a_ab,
            doses_b_ab,
            observed_responses_ab,
            est_theoretical_responses_ab,
        ),
    )


def model_1_param(
    gamma, doses_a, doses_b, observed_responses_ab, theoretical_responses_ab
):
    gamma_0 = gamma[0]

    residuals = gamma_0 - observed_responses_ab + theoretical_responses_ab

    return residuals


def model_4_param(
    gamma, doses_a, doses_b, observed_responses_ab, theoretical_responses_ab
):
    gamma_0, gamma_1, gamma_2, gamma_3 = gamma

    residuals = (
        gamma_0
        + gamma_1 * doses_a
        + gamma_2 * doses_b
        + gamma_3 * doses_a * doses_b
        - observed_responses_ab
        + theoretical_responses_ab
    )

    return residuals


def model_6_param(
    gamma, doses_a, doses_b, observed_responses_ab, theoretical_responses_ab
):
    gamma_0, gamma_1, gamma_2, gamma_3, gamma_4, gamma_5 = gamma

    residuals = (
        gamma_0
        + gamma_1 * doses_a
        + gamma_2 * doses_b
        + gamma_3 * doses_a * doses_b
        + gamma_4 * doses_a**2
        + gamma_5 * doses_b**2
        - observed_responses_ab
        + theoretical_responses_ab
    )

    return residuals


# convert values to % inhibition
def normalize(values, maximum=100, minimum=0):
    return 1 - (values - minimum) / (maximum - minimum)


def plot_heatmap(
    name_a,
    name_b,
    units_a,
    units_b,
    doses_a,
    doses_b,
    responses_a,
    responses_b,
    doses_a_ab,
    doses_b_ab,
    responses_ab,
    I_estimates,
    I_ci_his,
    I_ci_los,
    model_size,
    file_name_context=None,
):
    label_a = f"{name_a} Concentration ({units_a})"
    label_b = f"{name_b} Concentration ({units_a})"

    if file_name_context:
        file_name_context += "_"

    dose_response_dict = {x: y for x, y in zip(doses_a, responses_a)}
    dose_response_dict.update({x: y for x, y in zip(doses_b, responses_b)})

    data_dict = {}
    data_dict[label_a] = [float(x) for x in doses_a]
    data_dict[label_b] = [0] * len(doses_a)
    data_dict["I"] = [np.nan for y in responses_a]
    data_dict["M(I)"] = [np.nan for y in responses_a]
    data_dict["CI(I, lo)"] = [np.nan for y in responses_a]
    data_dict["CI(I, hi)"] = [np.nan for y in responses_a]

    data_dict[label_a] = np.append(data_dict[label_a], [0] * len(doses_b))
    data_dict[label_b] = np.append(data_dict[label_b], [float(x) for x in doses_b])
    data_dict["I"] = np.append(data_dict["I"], [np.nan for y in responses_b])
    data_dict["M(I)"] = np.append(data_dict["M(I)"], [np.nan for y in responses_b])
    data_dict["CI(I, lo)"] = np.append(
        data_dict["CI(I, lo)"], [np.nan for y in responses_b]
    )
    data_dict["CI(I, hi)"] = np.append(
        data_dict["CI(I, hi)"], [np.nan for y in responses_b]
    )

    for dose_a, dose_b, response_ab, I_est, I_ci_lower, I_ci_upper in zip(
        doses_a_ab, doses_b_ab, responses_ab, I_estimates, I_ci_los, I_ci_his
    ):
        data_dict[label_a] = np.append(data_dict[label_a], [float(dose_a)])
        data_dict[label_b] = np.append(data_dict[label_b], [float(dose_b)])

        response_a = dose_response_dict[dose_a]
        response_b = dose_response_dict[dose_b]
        response_ab_expected = response_a + response_b - response_a * response_b

        data_dict["I"] = np.append(data_dict["I"], [response_ab - response_ab_expected])
        data_dict["M(I)"] = np.append(data_dict["M(I)"], I_est)
        data_dict["CI(I, lo)"] = np.append(data_dict["CI(I, lo)"], I_ci_lower)
        data_dict["CI(I, hi)"] = np.append(data_dict["CI(I, hi)"], I_ci_upper)

    data = pd.DataFrame(data_dict)
    data["significance"] = data.apply(
        lambda row: "S" if row["CI(I, lo)"] * row["CI(I, hi)"] > 0 else "NS", axis=1
    )
    data["label"] = data.apply(lambda row: row2label(row), axis=1)

    data_values = data.pivot_table(
        index=label_a, columns=label_b, values="I", aggfunc="mean", dropna=False
    )
    data_annotations = data.pivot_table(
        index=label_a, columns=label_b, values="label", aggfunc="first", dropna=False
    )

    _fig = plt.figure(figsize=(12, 8), dpi=100)
    ax = sns.heatmap(
        data_values,
        vmin=-1,
        vmax=1,
        cmap="vlag_r",
        center=0,
        annot=data_annotations,
        fmt="",
        linewidths=2,
        square=True,
        cbar_kws={
            "extend": "both",
            "label": "Excess Over Bliss",
            # 'location': 'bottom',
            "ticks": [-1, 0, 1],
        },
    )
    ax.collections[0].colorbar.set_ticklabels(
        ["-1 (Antagonism)", "0 (Additivity)", "+1 (Synergy)"]
    )
    ax.invert_yaxis()
    plt.title(f"{name_a} vs. {name_b}: Bliss Ixn ({model_size}-param)")
    plt.tight_layout()
    uniq_str = str(int(time() * 1000) % 1_620_000_000_000)
    log_dir = f'{get_config_setting("log_dir")}/interactions2'
    os.makedirs(log_dir, exist_ok=True)
    plt.savefig(
        f"{log_dir}/{name_a}-{name_b}_{file_name_context}bliss_{model_size}-param_{uniq_str}.png"
    )
    plt.clf()


def print_gamma_table(gammas, gamma_ci_his, gamma_ci_los, model_size):
    parameters = ["γ₀", "γ₁", "γ₂", "γ₃", "γ₄", "γ₅"]
    parameters = parameters[:model_size]  # trim to appropriate size
    variables = ["", "a", "b", "ab", "a²", "b²"]
    variables = variables[:model_size]  # trim to appropriate size

    print(
        "Model: I =",
        " + ".join([param + var for param, var in zip(parameters, variables)]),
    )

    data = pd.DataFrame(
        {
            "Variable": variables,
            "Parameter": parameters,
            "Value": gammas,
            "CI (high)": gamma_ci_his,
            "CI (low)": gamma_ci_los,
            "Significant": [
                ci_lo * ci_hi > 0 for ci_lo, ci_hi in zip(gamma_ci_los, gamma_ci_his)
            ],
        }
    )

    print(data.transpose())


def print_mean(
    doses_a, doses_b, responses_a, responses_b, doses_a_ab, doses_b_ab, responses_ab
):
    dose_response_dict = {x: y for x, y in zip(doses_a, responses_a)}
    dose_response_dict.update({x: y for x, y in zip(doses_b, responses_b)})

    interaction_scores = []

    for dose_a, dose_b, response_ab in zip(doses_a_ab, doses_b_ab, responses_ab):
        response_a = dose_response_dict[dose_a]
        response_b = dose_response_dict[dose_b]
        response_ab_expected = response_a + response_b - response_a * response_b

        interaction_scores.append(response_ab - response_ab_expected)

    print("Mean interaction score:", np.nanmean(interaction_scores))


# as per formulas in Zhao 2014, https://doi.org/10.1177/1087057114521867
def response_surface(
    doses_a,
    responses_all_a,
    doses_b,
    responses_all_b,
    doses_a_ab,
    doses_b_ab,
    responses_all_ab,
    positive_control,
    sampling_iterations=1000,
    sample_size=20,
    model_size=4,
    alpha=0.05,
    file_name_context=None,
):
    positive_control_value = np.nanmean(positive_control)
    responses_all_a = normalize(
        responses_all_a, maximum=100, minimum=positive_control_value
    )
    responses_all_b = normalize(
        responses_all_b, maximum=100, minimum=positive_control_value
    )
    responses_all_ab = normalize(
        responses_all_ab, maximum=100, minimum=positive_control_value
    )

    #
    # Establish a model
    #

    est_true_responses_a = np.nanmean(responses_all_a, axis=1)
    est_true_responses_b = np.nanmean(responses_all_b, axis=1)

    # might need to replace this with Zhao 2014, Eqs. 7 and 8?
    est_response_covarmat_a = np.ma.cov(
        np.ma.masked_invalid(responses_all_a)
    )  # covar ignoring NaN
    est_response_covarmat_b = np.ma.cov(
        np.ma.masked_invalid(responses_all_b)
    )  # covar ignoring NaN

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        # feels like the error/variance in responses_all_ab should be taken into account?
        observed_responses_ab = np.nanmean(responses_all_ab, axis=1)
    valid_combo_idxs = ~np.isnan(observed_responses_ab)

    doses_a_ab = doses_a_ab[valid_combo_idxs]
    doses_b_ab = doses_b_ab[valid_combo_idxs]
    observed_responses_ab = observed_responses_ab[valid_combo_idxs]

    if model_size == 4:
        model_function = model_4_param
        gamma_guess_0 = [0.5, 0, 0, 0]
    elif model_size == 6:
        model_function = model_6_param
        gamma_guess_0 = [0.5, 0, 0, 0, 0, 0]
    elif model_size == 1:
        model_function = model_1_param
        gamma_guess_0 = [0]
    else:
        raise ValueError(f"Model with {model_size} parameters is not defined")

    gamma_sample_means = np.zeros((model_size, sampling_iterations))
    gamma_sample_covars = np.zeros((model_size, model_size, sampling_iterations))

    for sample_i in range(sampling_iterations):
        gammas = np.zeros((model_size, sample_size))
        for point_j in range(sample_size):
            model_j = fit_model_with_noise(
                model_function,
                gamma_guess_0,
                doses_a,
                doses_b,
                doses_a_ab,
                doses_b_ab,
                est_response_covarmat_a,
                est_response_covarmat_b,
                est_true_responses_a,
                est_true_responses_b,
                observed_responses_ab,
            )

            gammas[:, point_j] = model_j.x

        gamma_sample_means[:, sample_i] = np.mean(gammas, axis=1)
        gamma_sample_covars[:, :, sample_i] = np.cov(gammas)

    est_gamma = np.mean(gamma_sample_means, axis=1)
    est_gamma_covarmat = np.mean(gamma_sample_covars, axis=2) + np.cov(
        gamma_sample_means
    )

    #
    # Use the model (mainly `est_gamma`) to predict I and CI(I)
    #

    z = scipy.stats.norm.ppf(1 - alpha / 2)

    est_gamma_stddev = np.sqrt(np.diagonal(est_gamma_covarmat))

    interaction_index_estimates = []
    interaction_index_ci_his = []
    interaction_index_ci_los = []

    for dose_a, dose_b in zip(doses_a_ab, doses_b_ab):
        dose_a, dose_b = float(dose_a), float(dose_b)
        x = np.array([1, dose_a, dose_b, dose_a * dose_b, dose_a**2, dose_b**2])
        x = x[:model_size]  # trim to appropriate size

        interaction_index_estimate = np.dot(x, est_gamma)
        interaction_index_uncertainty = z * np.sqrt(
            x @ est_gamma_covarmat @ x
        )  # @ is matrix multiplication
        interaction_index_estimates.append(interaction_index_estimate)
        interaction_index_ci_his.append(
            interaction_index_estimate + interaction_index_uncertainty
        )
        interaction_index_ci_los.append(
            interaction_index_estimate - interaction_index_uncertainty
        )

    dose_a_0 = doses_a[0]
    dose_b_0 = doses_b[0]

    plot_heatmap(
        dose_a_0.drug,
        dose_b_0.drug,
        dose_a_0.unit,
        dose_b_0.unit,
        doses_a,
        doses_b,
        est_true_responses_a,
        est_true_responses_b,
        doses_a_ab,
        doses_b_ab,
        observed_responses_ab,
        interaction_index_estimates,
        interaction_index_ci_his,
        interaction_index_ci_los,
        model_size,
        file_name_context=file_name_context,
    )
    print_gamma_table(
        est_gamma,
        est_gamma + est_gamma_stddev * z,
        est_gamma - est_gamma_stddev * z,
        model_size,
    )
    print_mean(
        doses_a,
        doses_b,
        est_true_responses_a,
        est_true_responses_b,
        doses_a_ab,
        doses_b_ab,
        observed_responses_ab,
    )


def row2label(row):
    if np.isnan(row["I"]):
        return ""
    else:
        return "I = {:.2f}\nM(I) = {:.2f}\n({:.2f}, {:.2f})\n{}".format(
            row["I"],
            row["M(I)"],
            row["CI(I, lo)"],
            row["CI(I, hi)"],
            row["significance"],
        )


if __name__ == "__main__":
    ## example/test values

    doses_a = np.array([22.5, 45, 90, 180])
    responses_all_a = np.array(
        [
            [109.0, 108.0, 99.0, 60.0, 108.0, 121.0, np.nan],
            [76.0, 80.0, 101.0, 81.0, 106.0, 102.0, 85.0],
            [61.0, 74.0, 57.0, 75.0, 84.0, 89.0, 55.0],
            [16.0, 32.0, 82.0, 30.0, 41.0, 11.0, 24.0],
        ]
    )
    doses_b = np.array([1.7, 3.4, 6.8, 13.6])
    responses_all_b = np.array(
        [
            [97.0, 87.0, 112.0, 106.0, 99.0, 75.0, np.nan],
            [95.0, 62.0, 68.0, 70.0, 28.0, 68.0, 46.0],
            [25.0, 46.0, 40.0, 16.0, 25.0, 36.0, 48.0],
            [29.0, 14.0, 15.0, 22.0, 30.0, 30.0, 20.0],
        ]
    )
    doses_a_ab = np.array(
        [22.5, 45, 90, 180, 22.5, 45, 90, 180, 22.5, 45, 90, 180, 22.5, 45, 90, 180]
    )
    doses_b_ab = np.array(
        [
            1.7,
            1.7,
            1.7,
            1.7,
            3.4,
            3.4,
            3.4,
            3.4,
            6.8,
            6.8,
            6.8,
            6.8,
            13.6,
            13.6,
            13.6,
            13.6,
        ]
    )
    responses_all_ab = np.array(
        [
            [89.0, 115.0, 86.0, 74.0],
            [66.0, 92.0, 48.0, 59.0],
            [25.0, 46.0, 63.0, 47.0],
            [47.0, 20.0, 19.0, 24.0],
            [55.0, 86.0, 40.0, 50.0],
            [23.0, 64.0, 63.0, 16.0],
            [37.0, 45.0, 67.0, 34.0],
            [15.0, 20.0, 20.0, 29.0],
            [50.0, 41.0, 58.0, 11.0],
            [25.0, 14.0, 25.0, 27.0],
            [56.0, 50.0, 26.0, 29.0],
            [np.nan, np.nan, np.nan, np.nan],
            [23.0, 28.0, 38.0, 39.0],
            [28.0, 31.0, 8.0, 7.0],
            [13.0, 17.0, 42.0, 11.0],
            [np.nan, np.nan, np.nan, np.nan],
        ]
    )
    positive_control = np.array([10.0, 6.0, 6.0, 5.0])

    response_surface(
        doses_a,
        responses_all_a,
        doses_b,
        responses_all_b,
        doses_a_ab,
        doses_b_ab,
        responses_all_ab,
        positive_control,
        model_size=1,
    )
    response_surface(
        doses_a,
        responses_all_a,
        doses_b,
        responses_all_b,
        doses_a_ab,
        doses_b_ab,
        responses_all_ab,
        positive_control,
        model_size=4,
    )
    response_surface(
        doses_a,
        responses_all_a,
        doses_b,
        responses_all_b,
        doses_a_ab,
        doses_b_ab,
        responses_all_ab,
        positive_control,
        model_size=6,
    )



# Standard Library Imports
import unittest

# External Imports
import numpy as np

# Local Imports
from pepitatools.imageops import _get_bit_depth


class TestBitDepth(unittest.TestCase):
    def test_get_bit_depth(self):
        self.assertTupleEqual(
            _get_bit_depth(np.array([1, 2, 3, 4, 5])), (np.uint8, 255)
        )
        self.assertTupleEqual(
            _get_bit_depth(np.array([1, 2, 3, 4, 255])), (np.uint8, 255)
        )
        self.assertTupleEqual(
            _get_bit_depth(np.array([1, 2, 3, 4, 256])), (np.uint16, 65_535)
        )
        self.assertTupleEqual(
            _get_bit_depth(np.array([1, 2, 3, 4, 65_536])), (np.int32, 2_147_483_647)
        )


if __name__ == "__main__":
    unittest.main()



"""
Analyze plates
"""

# Imports
# Standard Library Imports
from __future__ import annotations
import csv
import os
import re
import sys
import warnings


# External Imports
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

# Local Imports
from . import imageops
from . import keyence
from .configuration import get_config_setting

# for windows consoles (e.g. git bash) to work properly
try:
    sys.stdin.reconfigure(encoding="utf-8")
    sys.stdout.reconfigure(encoding="utf-8")
except AttributeError:
    pass  # sys.stdout has been wrapped, but should already support utf-8


class Image:
    def __init__(
        self,
        filename,
        group,
        debug=0,
    ):
        # Get values from config
        self.channel = int(get_config_setting("channel_main_ototox"))
        self.channel_subtr = int(get_config_setting("channel_subtr_ototox"))
        self.particles = True
        self.replacement_delim = get_config_setting("filename_replacement_delimiter")
        self.replacement_brfld = get_config_setting(
            "filename_replacement_brightfield_ototox"
        ).split(self.replacement_delim)
        self.replacement_mask = get_config_setting(
            "filename_replacement_mask_ototox"
        ).split(self.replacement_delim)
        self.replacement_subtr = get_config_setting(
            "filename_replacement_subtr_ototox"
        ).split(self.replacement_delim)
        self.fl_filename = filename
        self.bf_filename = filename.replace(
            self.replacement_brfld[0], self.replacement_brfld[1]
        )
        self.subtr_filename = filename.replace(
            self.replacement_subtr[0], self.replacement_subtr[1]
        )

        match = re.search(r"([a-zA-Z0-9]+)_XY([0-9][0-9])_", filename)
        if not match:
            raise UserError("Filename %s missing needed xy information" % filename)

        self.plate = match.group(1)
        self.xy = int(match.group(2))

        self.group = group
        self.debug = debug

        self.bf_img = None
        self.bf_metadata = None
        self.fl_img = None
        self.fl_metadata = None
        self.subtr_img = None
        self.mask = None
        self.normalized_value = None
        self.value = None

    def get_bf_img(self):
        if self.bf_img is None:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                self.bf_img = imageops.read(self.bf_filename, np.uint16)
        return self.bf_img

    def get_bf_metadata(self):
        if self.bf_metadata is None:
            self.bf_metadata = keyence.extract_metadata(self.bf_filename)
        return self.bf_metadata

    def get_fl_img(self):
        if self.fl_img is None:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                self.fl_img = imageops.read(self.fl_filename, np.uint16, self.channel)
        return self.fl_img

    def get_fl_metadata(self):
        if self.fl_metadata is None:
            self.fl_metadata = keyence.extract_metadata(self.fl_filename)
        return self.fl_metadata

    def get_mask(self):
        if self.mask is None:
            self.mask = imageops.get_fish_mask(
                self.get_bf_img(),
                self.get_fl_img(),
                particles=self.particles,
                silent=self.debug < 1,
                verbose=self.debug >= 2,
                v_file_prefix="{}_XY{:02d}".format(self.plate, self.xy),
                mask_filename=self.fl_filename.replace(
                    self.replacement_mask[0], self.replacement_mask[1]
                ),
                subtr_img=self.get_subtr_img(),
            )
        return self.mask

    def get_raw_value(self):
        if self.value is None:
            fl_img_masked = imageops.apply_mask(self.get_fl_img(), self.get_mask())
            score = imageops.score(fl_img_masked)
            self.value = score if score > 0 else np.nan
        return self.value

    def get_subtr_img(self):
        if self.subtr_img is None:
            if os.path.isfile(self.subtr_filename):
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", UserWarning)
                    self.subtr_img = imageops.read(
                        self.subtr_filename, np.uint16, self.channel_subtr
                    )
            else:
                self.subtr_img = []
        return self.subtr_img

    def normalize(self, control_values, cap):
        try:
            val = float(self.get_raw_value() * 100 // control_values[self.plate])
            if cap > 0:
                self.normalized_value = val if val < cap else np.nan
            else:
                self.normalized_value = val
        except ZeroDivisionError:
            print(
                "ERROR: Plate",
                self.plate,
                "group",
                self.group,
                "with value",
                self.get_raw_value(),
                "has control value",
                control_values[self.plate],
            )
            self.normalized_value = np.nan
        return self


class UserError(ValueError):
    pass


def chart(results, chartfile, scale="linear"):
    with sns.axes_style(style="whitegrid"):
        data = pd.DataFrame(
            {
                "brightness": [
                    value for values in results.values() for value in values
                ],
                "group": [key for key, values in results.items() for _ in values],
            }
        )

        _fig = plt.figure(figsize=(12, 8), dpi=100)
        ax = sns.swarmplot(x="group", y="brightness", data=data)
        ax.set_yscale(scale)
        if scale == "linear":
            ax.set_ylim(bottom=0)
        sns.boxplot(
            x="group",
            y="brightness",
            data=data,
            showbox=False,
            showcaps=False,
            showfliers=False,
            whiskerprops={"visible": False},
        )
        plt.xticks(rotation=90)
        plt.tight_layout()
        plt.savefig(chartfile)


def get_schematic(platefile, target_count, plate_ignore=None, flat=True):
    if not platefile:
        return keyence.LAYOUT_DEFAULT
    if plate_ignore is None:
        plate_ignore = []

    if "" not in plate_ignore:
        plate_ignore.append("")

    with open(platefile, encoding="utf8", newline="") as f:
        schematic = [
            [well for well in row if well not in plate_ignore] for row in csv.reader(f)
        ]

    count = sum([len(row) for row in schematic])
    if (
        count != target_count
    ):  # try removing first row and first column, see if then it matches up
        del schematic[0]
        for row in schematic:
            del row[0]
        count = sum([len(row) for row in schematic])
        if count != target_count:
            raise UserError(
                f"Schematic does not have same number of cells ({count}) as images provided "
                + f"({target_count})"
            )

    return schematic if not flat else [well for row in schematic for well in row]


def quantify(
    imagefiles, plate_control=None, cap=-1, debug=0, group_regex=".*", schematic=None
):
    if plate_control is None:
        plate_control = ["B"]
    pattern = re.compile(group_regex)
    images = [
        Image(filename, group, debug)
        for filename, group in zip(imagefiles, schematic)
        if group in plate_control or pattern.search(group)
    ]
    control_values = _calculate_control_values(images, plate_control)
    return [image.normalize(control_values, cap) for image in images]


def _calculate_control_values(images, plate_control):
    ctrl_imgs = [img for img in images if img.group in plate_control]
    ctrl_vals = {}

    for plate in np.unique([img.plate for img in ctrl_imgs]):
        ctrl_results = np.array(
            [img.get_raw_value() for img in ctrl_imgs if img.plate == plate]
        )
        ctrl_vals[plate] = float(np.nanmedian(ctrl_results))

    if not ctrl_vals:
        raise UserError(
            "No control wells found. Please supply a --plate-control, or modify the given value."
        )

    return ctrl_vals


def main(
    imagefiles,
    cap=-1,
    chartfile=None,
    debug=0,
    group_regex=".*",
    platefile=None,
    plate_control=None,
    plate_ignore=None,
    silent=False,
):
    if plate_control is None:
        plate_control = ["B"]
    if plate_ignore is None:
        plate_ignore = []
    results = {}

    schematic = get_schematic(platefile, len(imagefiles), plate_ignore)
    groups = list(dict.fromkeys(schematic))  # deduplicated copy of `schematic`
    images = quantify(
        imagefiles,
        plate_control,
        cap=cap,
        debug=debug,
        group_regex=group_regex,
        schematic=schematic,
    )

    pattern = re.compile(group_regex)
    for group in groups:
        if group in plate_control or pattern.search(group):
            relevant_values = [
                img.normalized_value for img in images if img.group == group
            ]
            results[group] = relevant_values
            if not silent:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", RuntimeWarning)
                    print(group, np.nanmedian(relevant_values), relevant_values)

    if chartfile:
        chart(results, chartfile)

    return results



""" """

# Imports
# Standard Library Imports
from __future__ import annotations
import math
import re
import warnings

# External Imports
import numpy as np

# Local Imports
from . import analyze


# from ISO 12232:1998 via https://en.wikipedia.org/wiki/Film_speed#Measurements_and_calculations
# H = qLt/(N^2)
# H: luminous exposure (lux-seconds), proportional to pixel value
# L: luminance = variable of interest (candela/m^2)
# t: exposure time (seconds)
# N: aperture f-stop (unitless)
# q = (π/4) * T * v(θ) * cos^4(θ)
# T: transmittance of lens system (should be roughly constant for a given microscope)
# v: vignetting factor (currently ignored, could maybe be calculated from brightfield image)
# θ: angle relative to the lens
# therefore solved for L: L = H * N^2 / qt
# from https://en.wikipedia.org/wiki/Numerical_aperture#Numerical_aperture_versus_f-number
# N = f/D; NA = n * sin(arctan(D/2f))
# f: focal length of optical system
# D: diameter of entrance pupil
# NA: numerical aperture of the lens
# n: index of refraction of working medium (1.00 for air, 1.33 for water, 1.52 for immersion oil)
# therefore N in terms of NA and n: N = 1 / (2 * tan(arcsin(NA/n)))
# values are returned in units proportional to candela/m^2
def get_absolute_value(
    image, debug=0, n=1.0, transmittance=1, vignette=lambda theta: 1
):
    metadata = image.get_fl_metadata()

    H = image.get_raw_value()
    N = 1 / (2 * math.tan(math.asin(metadata["Numerical Aperture"] / n)))
    theta = 0  # TODO: calculate theta from working distance & weighted average(?) of signal in img
    q = (math.pi / 4) * transmittance * vignette(theta) * math.cos(theta) ** 4
    t = metadata["Exposure"]["Value"]

    if debug >= 1:
        print(
            '%s: H=%f "lx⋅s", N=%f, theta=%f rad, q=%f, t=%fs'
            % (image.fl_filename, H, N, theta, q, t)
        )

    L = H * N**2 / q / t

    return (
        np.nan if np.isnan(L) else int(L / 10_000)
    )  # to make results more readable, "cd/cm^2"


def main(
    imagefiles,
    cap=-1,
    chartfile=None,
    debug=0,
    group_regex=".*",
    platefile=None,
    plate_control=["B"],
    plate_ignore=[],
    silent=False,
):
    results = {}

    schematic = analyze.get_schematic(platefile, len(imagefiles), plate_ignore)
    groups = list(dict.fromkeys(schematic))  # deduplicated copy of `schematic`
    pattern = re.compile(group_regex)
    images = [
        analyze.Image(filename, group, debug)
        for filename, group in zip(imagefiles, schematic)
        if group in plate_control or pattern.search(group)
    ]

    pattern = re.compile(group_regex)
    for group in groups:
        if group in plate_control or pattern.search(group):
            relevant_values = [
                get_absolute_value(img) for img in images if img.group == group
            ]
            results[group] = relevant_values
            if not silent:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", RuntimeWarning)
                    print(group, np.nanmedian(relevant_values), relevant_values)

    if chartfile:
        analyze.chart(results, chartfile)

    return results



"""
Provide global configuration
"""

# Imports
# Standard Library Imports
import configparser
import os
import pathlib
from typing import Union

# External Imports

# Local Imports

__all__ = ["get_config_setting", "set_config_setting", "read_config"]

_CONFIG_SECTION = "Main"

# Set Default Configuration Options
_CONFIGURATION = {
    "absolute_max_infection": 26249,
    "absolute_min_infection": 431,
    "absolute_max_ototox": 26249,
    "absolute_min_ototox": 431,
    "channel_main_ototox": 1,
    "channel_main_infection": 0,
    "channel_subtr_ototox": 0,
    "channel_subtr_infection": 1,
    "filename_replacement_delimiter": "|",
    "filename_replacement_brightfield_infection": "CH2|CH4",
    "filename_replacement_brightfield_ototox": "CH1|CH4",
    "filename_replacement_mask_infection": "CH2|mask",
    "filename_replacement_mask_ototox": "CH1|mask",
    "filename_replacement_subtr_infection": "CH2|CH1",
    "filename_replacement_subtr_ototox": "CH1|CH2",
    "log_dir": "/path/to/log/dir",
    "keyence_lenses_file": None,
}

_DEFAULT_CONFIG_STR = """
[Main]
absolute_max_infection = 26249
absolute_min_infection = 431
absolute_max_ototox = 26249
absolute_min_ototox = 431
channel_main_ototox = 1
channel_main_infection = 0
channel_subtr_ototox = 0
channel_subtr_infection = 1
filename_replacement_delimiter = |
filename_replacement_brightfield_infection = CH2|CH4
filename_replacement_brightfield_ototox = CH1|CH4
filename_replacement_mask_infection = CH2|mask
filename_replacement_mask_ototox = CH1|mask
filename_replacement_subtr_infection = CH2|CH1
filename_replacement_subtr_ototox = CH1|CH2
log_dir = /path/to/log/dir
keyence_lenses_file = /path/to/keyence/file
"""


def get_config_setting(setting):
    return _CONFIGURATION.get(setting, None)


def set_config_setting(setting, value):
    _CONFIGURATION[setting] = value


def read_config(config_file: Union[str, os.PathLike, None] = None) -> None:
    if config_file is None:
        config_file = pathlib.Path(".") / "config.ini"
    else:
        config_file = pathlib.Path(config_file)
    # If the config file isn't created, keep going to try and use config options passed in from the
    # command line
    if not config_file.exists():
        return None

    parsed_config = configparser.ConfigParser()
    parsed_config.read(str(config_file))

    # Update the config, falling back on the default
    global _CONFIGURATION
    _CONFIGURATION["absolute_max_infection"] = parsed_config.get(
        _CONFIG_SECTION, "absolute_max_infection", fallback=26249
    )
    _CONFIGURATION["absolute_min_infection"] = parsed_config.get(
        _CONFIG_SECTION, "absolute_min_infection", fallback=431
    )
    _CONFIGURATION["absolute_max_ototox"] = parsed_config.get(
        _CONFIG_SECTION, "absolute_max_ototox", fallback=26249
    )
    _CONFIGURATION["absolute_min_ototox"] = parsed_config.get(
        _CONFIG_SECTION, "absolute_min_ototox", fallback=431
    )
    _CONFIGURATION["channel_main_ototox"] = parsed_config.get(
        _CONFIG_SECTION, "channel_main_ototox", fallback=1
    )
    _CONFIGURATION["channel_main_infection"] = parsed_config.get(
        _CONFIG_SECTION, "channel_main_infection", fallback=0
    )
    _CONFIGURATION["channel_subtr_ototox"] = parsed_config.get(
        _CONFIG_SECTION, "channel_subtr_ototox", fallback=0
    )
    _CONFIGURATION["channel_subtr_infection"] = parsed_config.get(
        _CONFIG_SECTION, "channel_subtr_infection", fallback=1
    )
    _CONFIGURATION["filename_replacement_delimiter"] = parsed_config.get(
        _CONFIG_SECTION, "filename_replacement_delimiter", fallback="|"
    )
    _CONFIGURATION["filename_replacement_brightfield_infection"] = parsed_config.get(
        _CONFIG_SECTION,
        "filename_replacement_brightfield_infection",
        fallback="CH2|CH4",
    )
    _CONFIGURATION["filename_replacement_brightfield_ototox"] = parsed_config.get(
        _CONFIG_SECTION, "filename_replacement_brightfield_ototox", fallback="CH1|CH4"
    )
    _CONFIGURATION["filename_replacement_mask_infection"] = parsed_config.get(
        _CONFIG_SECTION, "filename_replacement_mask_infection", fallback="CH2|mask"
    )
    _CONFIGURATION["filename_replacement_mask_ototox"] = parsed_config.get(
        _CONFIG_SECTION, "filename_replacement_mask_ototox", fallback="CH1|mask"
    )
    _CONFIGURATION["filename_replacement_subtr_infection"] = parsed_config.get(
        _CONFIG_SECTION, "filename_replacement_subtr_infection", fallback="CH2|CH1"
    )
    _CONFIGURATION["filename_replacement_subtr_ototox"] = parsed_config.get(
        _CONFIG_SECTION, "filename_replacement_subtr_ototox", fallback="CH1|CH2"
    )
    _CONFIGURATION["log_dir"] = parsed_config.get(
        _CONFIG_SECTION, "log_dir", fallback="/path/to/log/dir"
    )
    _CONFIGURATION["keyence_lenses_file"] = parsed_config.get(
        _CONFIG_SECTION, "keyence_file", fallback=None
    )

    # Make sure the keyence_file isn't just the default, if it is set it to None for easier handling elsewhere
    if _CONFIGURATION["keyence_lenses_file"] == "/path/to/keyence/file":
        _CONFIGURATION["keyence_lenses_file"] = None



# Imports
# Standard Library Imports
import pathlib
import unittest

# External Imports
from pepitatools.configuration import set_config_setting

# Local Imports
from pepitatools import dose_response, utils


class TestDoseResponse(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.log_dir = pathlib.Path(__file__).parent / "tmp_logs"
        cls.log_dir.mkdir(exist_ok=True)
        set_config_setting("log_dir", str(cls.log_dir))

    @classmethod
    def tearDownClass(cls):
        for f in cls.log_dir.glob("*"):
            if f.name == "dose_response":
                for sub_f in f.glob("*"):
                    sub_f.unlink()
                f.rmdir()
                continue
            f.unlink()
        cls.log_dir.rmdir()

    def test_do_additive_isobole(self):
        self.assertTrue(
            utils.equalsish(
                0.46255,
                dose_response.do_additive_isobole(
                    a_i=25,
                    A_E50_a=65.8,
                    B_E50_b=3.99,
                    E_max_a=1.58,
                    E_max_b=4.17,
                    B_i=1.2,
                    p=1.73,
                    q=1.92,
                ),
            )
        )
        self.assertTrue(
            utils.equalsish(
                2.06255,
                dose_response.do_additive_isobole(
                    a_i=25,
                    A_E50_a=65.8,
                    B_E50_b=3.99,
                    E_max_a=1.58,
                    E_max_b=4.17,
                    B_i=2.8,
                    p=1.73,
                    q=1.92,
                ),
            )
        )

        self.assertTrue(
            utils.equalsish(
                4.46255,
                dose_response.do_additive_isobole(
                    a_i=25,
                    A_E50_a=65.8,
                    B_E50_b=3.99,
                    E_max_a=1.58,
                    E_max_b=4.17,
                    B_i=5.2,
                    p=1.73,
                    q=1.92,
                ),
            )
        )

        self.assertTrue(
            utils.equalsish(
                0.60894,
                dose_response.do_additive_isobole(
                    a_i=100,
                    A_E50_a=65.8,
                    B_E50_b=3.99,
                    E_max_a=1.58,
                    E_max_b=4.17,
                    B_i=2.8,
                    p=1.73,
                    q=1.92,
                ),
            )
        )

        self.assertTrue(
            utils.equalsish(
                3.00894,
                dose_response.do_additive_isobole(
                    a_i=100,
                    A_E50_a=65.8,
                    B_E50_b=3.99,
                    E_max_a=1.58,
                    E_max_b=4.17,
                    B_i=5.2,
                    p=1.73,
                    q=1.92,
                ),
            )
        )

        self.assertTrue(
            utils.equalsish(
                2.28538,
                dose_response.do_additive_isobole(
                    a_i=400,
                    A_E50_a=65.8,
                    B_E50_b=3.99,
                    E_max_a=1.58,
                    E_max_b=4.17,
                    B_i=5.2,
                    p=1.73,
                    q=1.92,
                ),
            )
        )

    def test_do_fic(self):
        self.assertTrue(
            utils.equalsish(
                1,
                dose_response.do_FIC(
                    a_i=25,
                    b_i=0.46255,
                    A_E50_a=65.8,
                    B_E50_b=3.99,
                    E_max_a=1.58,
                    E_max_b=4.17,
                    B_i=1.2,
                    p=1.73,
                    q=1.92,
                ),
            )
        )

        self.assertTrue(
            utils.equalsish(
                1,
                dose_response.do_FIC(
                    a_i=25,
                    b_i=2.06255,
                    A_E50_a=65.8,
                    B_E50_b=3.99,
                    E_max_a=1.58,
                    E_max_b=4.17,
                    B_i=2.8,
                    p=1.73,
                    q=1.92,
                ),
            )
        )

        self.assertTrue(
            utils.equalsish(
                1,
                dose_response.do_FIC(
                    a_i=25,
                    b_i=4.46255,
                    A_E50_a=65.8,
                    B_E50_b=3.99,
                    E_max_a=1.58,
                    E_max_b=4.17,
                    B_i=5.2,
                    p=1.73,
                    q=1.92,
                ),
            )
        )

        self.assertTrue(
            utils.equalsish(
                1,
                dose_response.do_FIC(
                    a_i=100,
                    b_i=0.60894,
                    A_E50_a=65.8,
                    B_E50_b=3.99,
                    E_max_a=1.58,
                    E_max_b=4.17,
                    B_i=2.8,
                    p=1.73,
                    q=1.92,
                ),
            )
        )

        self.assertTrue(
            utils.equalsish(
                1,
                dose_response.do_FIC(
                    a_i=100,
                    b_i=3.00894,
                    A_E50_a=65.8,
                    B_E50_b=3.99,
                    E_max_a=1.58,
                    E_max_b=4.17,
                    B_i=5.2,
                    p=1.73,
                    q=1.92,
                ),
            )
        )

        self.assertTrue(
            utils.equalsish(
                1,
                dose_response.do_FIC(
                    a_i=400,
                    b_i=2.28538,
                    A_E50_a=65.8,
                    B_E50_b=3.99,
                    E_max_a=1.58,
                    E_max_b=4.17,
                    B_i=5.2,
                    p=1.73,
                    q=1.92,
                ),
            )
        )

    def test_filter_valid(self):
        self.assertListEqual(
            dose_response.filter_valid([1, 1, 2, 3, 5, 8], minimum=3), [3, 5, 8]
        )
        self.assertListEqual(
            dose_response.filter_valid([1, 1, 2, 3, 5, 8], tolerance=2), [1, 3, 5, 8]
        )
        self.assertListEqual(
            dose_response.filter_valid([1, 1, 2, 3, 5, 8], tolerance=3), [1, 5, 8]
        )
        self.assertListEqual(
            dose_response.filter_valid([1, 1, 2, 3, 5, 8], minimum=3, tolerance=3),
            [
                3,
                8,
            ],
        )

    def test_model(self):
        model = dose_response._get_neo_model()
        self.assertLess(model.effective_concentration(0.001), 1)
        self.assertLess(model.get_absolute_E_max(), 50)
        self.assertLess(model.get_condition_E_max(), 50)
        self.assertTrue(utils.equalsish(1, model.get_pct_survival(xs=0.001)))
        self.assertTrue(
            utils.equalsish(0.0, model.get_pct_survival(xs=2000), delta=0.03)
        )  # THIS TEST IS CHANGED
        self.assertEqual(model.get_pct_survival(ys=100), 1)
        self.assertLessEqual(model.get_pct_survival(ys=0.001), 0)
        self.assertTrue(
            utils.equalsish(0, model.get_pct_survival(ys=model.get_absolute_E_max()))
        )
        self.assertEqual(model.get_x_units(), "μM")
        self.assertGreater(model.get_ys(0.001), 99)
        self.assertLess(model.get_ys(2000), 50)


if __name__ == "__main__":
    unittest.main()



"""
Main CLI for pepita-tools
"""

# Imports
# Standard library imports
from __future__ import annotations
import argparse
import json
import os
import sys
from time import time
import warnings

# External imports
from matplotlib import pyplot as plt
import numpy as np


# Local Imports
from pepitatools import (
    analyze,
    absolute,
    dose_response,
    imagej_scripts,
    imageops,
    infection,
    keyence,
    pipeline,
    utils,
)
from pepitatools.configuration import (
    read_config,
    get_config_setting,
    set_config_setting,
    _DEFAULT_CONFIG_STR,
)


def set_arguments(parser):
    parser.add_argument(
        "imagefiles",
        nargs="+",
        help="The absolute or relative filenames where the relevant images can be found.",
    )
    parser.add_argument(
        "-ch",
        "--chartfile",
        help="If supplied, the resulting numbers will be charted at the given filename.",
    )

    parser.add_argument(
        "-p",
        "--platefile",
        help="CSV file containing a schematic of the plate from which the given images were "
        "taken. Row and column headers are optional. The cell values are essentially just "
        "arbitrary labels: results will be grouped and charted according to the supplied "
        "values.",
    )
    parser.add_argument(
        "-pc",
        "--plate-control",
        default=["B"],
        nargs="*",
        help=(
            "Labels to treat as the control condition in the plate schematic. These wells are "
            "used to normalize all values in the plate for more interpretable results. Any number "
            "of values may be passed."
        ),
    )
    parser.add_argument(
        "-pi",
        "--plate-ignore",
        default=[],
        nargs="*",
        help=(
            "Labels to ignore (treat as null/empty) in the plate schematic. Empty cells will "
            'automatically be ignored, but any other null values (e.g. "[empty]") must be '
            "specified here. Any number of values may be passed."
        ),
    )

    parser.add_argument(
        "-g",
        "--group-regex",
        default=".*",
        help=(
            "Pattern to be used to match group names that should be included in the results. "
            "Matched groups will be included, groups that don't match will be ignored. Control "
            "wells will always be included regardless of whether they match."
        ),
    )

    parser.add_argument(
        "-c",
        "--cap",
        default=-1,
        type=int,
        help=(
            "Exclude well values larger than the given integer, expressed as a percentage of "
            "the median control value."
        ),
    )

    parser.add_argument(
        "-d",
        "--debug",
        action="count",
        default=0,
        help=(
            "Indicates intermediate processing images should be output for troubleshooting "
            "purposes. Including this argument once will yield one intermediate image per input "
            "file, twice will yield several intermediate images per input file."
        ),
    )
    parser.add_argument(
        "-s",
        "--silent",
        action="store_true",
        help=(
            "If present, printed output will be suppressed. More convenient for programmatic "
            "execution."
        ),
    )


# region Merge Config and Command Line Options
def merge_config_cli(args):
    """Take in command line arguments, and update the configuration based on them"""
    if args.absolute_max_infection:
        set_config_setting("absolute_max_infection", args.absolute_max_infection)
    if args.absolute_min_infection:
        set_config_setting("absolute_min_infection", args.absolute_min_infection)
    if args.absolute_max_ototox:
        set_config_setting("absolute_max_ototox", args.absolute_max_ototox)
    if args.absolute_min_ototox:
        set_config_setting("absolute_min_ototox", args.absolute_min_ototox)
    if args.channel_main_ototox:
        set_config_setting("channel_main_ototox", args.channel_main_ototox)
    if args.channel_main_infection:
        set_config_setting("channel_main_infection", args.channel_main_infection)
    if args.channel_subtr_ototox:
        set_config_setting("channel_subtr_ototox", args.channel_subtr_ototox)
    if args.channel_subtr_infection:
        set_config_setting("channel_subtr_infection", args.channel_subtr_infection)
    if args.filename_replacement_delimiter:
        set_config_setting(
            "filename_replacement_delimiter", args.filename_replacement_delimiter
        )
    if args.filename_replacement_brightfield_infection:
        set_config_setting(
            "filename_replacement_brightfield_infection",
            args.filename_replacement_brightfield_infection,
        )
    if args.filename_replacement_brightfield_ototox:
        set_config_setting(
            "filename_replacement_brightfield_ototox",
            args.filename_replacement_brightfield_ototox,
        )
    if args.filename_replacement_mask_infection:
        set_config_setting(
            "filename_replacement_mask_infection",
            args.filename_replacement_mask_infection,
        )
    if args.filename_replacement_mask_ototox:
        set_config_setting(
            "filename_replacement_mask_ototox", args.filename_replacement_mask_ototox
        )
    if args.filename_replacement_subtr_infection:
        set_config_setting(
            "filename_replacement_subtr_infection",
            args.filename_replacement_subtr_infection,
        )
    if args.filename_replacement_subtr_ototox:
        set_config_setting(
            "filename_replacement_subtr_ototox", args.filename_replacement_subtr_ototox
        )
    if args.log_dir:
        set_config_setting("log_dir", args.log_dir)
    if args.keyence_lenses_file:
        set_config_setting("keyence_lenses_file", args.keyence_lenses_file)


# endregion Merge Config and Command Line Options


# region subcommands
def config_file_command(args):
    if args.directory is None:
        directory = os.getcwd()
    else:
        directory = args.directory
    with open(os.path.join(directory, "config.ini"), "w") as f:
        f.write(_DEFAULT_CONFIG_STR)


def keyence_file_command(args):
    if args.directory is None:
        directory = os.getcwd()
    else:
        directory = args.directory
    keyence.write_example(os.path.join(directory, "keyence_lenses.csv"))


def absolute_command(args):
    args_dict = vars(args)
    absolute.main(**args_dict)


def analyze_command(args):
    args_dict = vars(args)
    analyze.main(**args_dict)


def keyence_command(args):
    for filename in args.filenames:
        metadata = keyence.extract_metadata(filename)
        print(filename, json.dumps(metadata, indent=2))


def imageops_command(args):
    for bf_filename in args.imagefiles:
        fl_filename = bf_filename.replace("CH4", "CH1")
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            bf_img = imageops.read(bf_filename, np.uint16)
            fl_img = (
                None if not args.particles else imageops.read(fl_filename, np.uint16, 1)
            )
        imageops.get_fish_mask(
            bf_img,
            fl_img,
            particles=args.particles,
            silent=args.debug < 1,
            verbose=args.debug > 1,
            v_file_prefix="imageops",
            mask_filename=bf_filename.replace("CH4", "mask"),
        )


def dose_response_command(args):
    if args.filename is None:
        models = [dose_response._get_neo_model(debug=0)]
    else:
        models = [dose_response._get_model(f) for f in args.filename]
    colors = ["#def", "#cdf", "#bcf", "#abf", "#9af", "#89f", "#78f"]

    for i, model in enumerate(models):
        print(model.cocktail)
        ec_90 = model.effective_concentration(0.9)
        ec_75 = model.effective_concentration(0.75)
        ec_50 = model.effective_concentration(0.5)

        print(f"E_max: {model.get_absolute_E_max()} score")
        print(f"EC_90: {ec_90} μM")
        print(f"ec_75: {ec_75} μM")
        print(f"ec_50: {ec_50} μM")

        model.chart(close=False, color=colors[i])

    plt.xlabel(f"{models[0].get_condition()} Dose (μM)")
    plt.ylabel("Pipeline Score")

    uniq_str = str(int(time() * 1000) % 1_620_000_000_000)
    plt.savefig(
        os.path.join(
            get_config_setting("log_dir"), f"{models[0].get_condition()}_{uniq_str}.png"
        )
    )
    plt.close()
    plt.clf()


def infection_command(args):
    args_dict = vars(args)
    try:
        infection.main(**args_dict)
    except analyze.UserError as ue:
        print("Error:", ue)
        sys.exit(1)


def pipeline_command(args):
    args_dict = vars(args)
    try:
        pipeline.main(**args_dict)
    except analyze.UserError as ue:
        print("Error:", ue)
        sys.exit(1)


def imagej_scripts_command(args):
    if args.scripts is None:
        scripts = [
            "openformasking",
            "openformaskingsingle",
            "savefishmask",
            "savefishnullmask",
            "macroize",
        ]
    else:
        scripts = args.scripts
    for script in scripts:
        imagej_scripts.write_script(script, args.directory)


# endregion subcommands


def create_parser():
    # region toplevel parser
    top_parser = argparse.ArgumentParser(prog="PEPITA-tools")
    top_parser.add_argument(
        "--config",
        dest="config",
        required=False,
        default="./config.ini",
        help="Path to config file, if not provided will look for config.ini in current "
        "directory",
    )

    # Add arguments which will be suppressed, but which can override the config file
    top_parser.add_argument(
        "--absolute_max_infection",
        dest="absolute_max_infection",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    top_parser.add_argument(
        "--absolute_min_infection",
        dest="absolute_min_infection",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    top_parser.add_argument(
        "--absolute_max_ototox",
        dest="absolute_max_ototox",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    top_parser.add_argument(
        "--absolute_min_ototox",
        dest="absolute_min_ototox",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    top_parser.add_argument(
        "--channel_main_ototox",
        dest="channel_main_ototox",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    top_parser.add_argument(
        "--channel_main_infection",
        dest="channel_main_infection",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    top_parser.add_argument(
        "--channel_subtr_ototox",
        dest="channel_subtr_ototox",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    top_parser.add_argument(
        "--channel_subtr_infection",
        dest="channel_subtr_infection",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    top_parser.add_argument(
        "--filename_replacement_delimiter",
        dest="filename_replacement_delimiter",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    top_parser.add_argument(
        "--filename_replacement_brightfield_infection",
        dest="filename_replacement_brightfield_infection",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    top_parser.add_argument(
        "--filename_replacement_brightfield_ototox",
        dest="filename_replacement_brightfield_ototox",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    top_parser.add_argument(
        "--filename_replacement_mask_infection",
        dest="filename_replacement_mask_infection",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    top_parser.add_argument(
        "--filename_replacement_mask_ototox",
        dest="filename_replacement_mask_ototox",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    top_parser.add_argument(
        "--filename_replacement_subtr_infection",
        dest="filename_replacement_subtr_infection",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    top_parser.add_argument(
        "--filename_replacement_subtr_ototox",
        dest="filename_replacement_subtr_ototox",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    top_parser.add_argument(
        "--log_dir",
        dest="log_dir",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    top_parser.add_argument(
        "--keyence_lenses_file",
        dest="keyence_lenses_file",
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )

    subparsers = top_parser.add_subparsers(help="Subcommands", required=True)
    top_parser.set_defaults(command=None)
    # endregion toplevel parser

    # region config-file parser
    # Create the parse for the "config-file" command
    config_file_parser = subparsers.add_parser(
        "config-file", help="Create a default config file"
    )
    config_file_parser.add_argument(
        "-d",
        "--directory",
        required=False,
        default=None,
        type=str,
        help="Directory to place default config file, current directory is used if not provided",
    )
    config_file_parser.set_defaults(func=config_file_command, command="config-file")
    # endregion config-file parser

    # region keyence-file parser
    keyence_file_parser = subparsers.add_parser(
        "keyence-file", help="Create a default keyence lenses file"
    )
    keyence_file_parser.add_argument(
        "-d",
        "--directory",
        required=False,
        default=None,
        type=str,
        help="Directory to place default config file, "
        "current directory is used if not provided",
    )
    keyence_file_parser.set_defaults(func=keyence_file_command, command="keyence-file")
    # endregion keyence-file parser

    # region absolute parser
    # Create the parser for the absolute script
    absolute_parser = subparsers.add_parser(
        "absolute",
        help="Analyzer for images of whole zebrafish with stained neuromasts, for the "
        "purposes of measuring hair cell damage in absolute terms. Reports values in "
        "arbitrary units not relative to any other value.",
    )
    set_arguments(absolute_parser)
    absolute_parser.set_defaults(func=absolute_command)
    # endregion absolute parser

    # region analyze parser
    # Create the parser for the Analyze Script
    analyze_parser = subparsers.add_parser(
        "analyze",
        help="Analyzer for images of whole zebrafish with stained "
        "neuromasts, for the purposes of measuring hair cell damage."
        " Reports values relative to control.",
    )
    set_arguments(analyze_parser)
    analyze_parser.set_defaults(func=analyze_command)
    # endregion analyze parser

    # region keyence
    # Create the Parser for the keyence script
    keyence_parser = subparsers.add_parser(
        "keyence", help="Print the metadata from keyence files"
    )
    keyence_parser.add_argument(
        "filenames", nargs="+", help="Files to get metadata from"
    )
    keyence_parser.set_defaults(func=keyence_command)

    # Create the parser for the imageops command
    imageops_parser = subparsers.add_parser(
        "imageops",
        help="Utility for operating on images of whole zebrafish with stained neuromasts, "
        "for the purposes of measuring hair cell damage.",
    )
    imageops_parser.add_argument(
        "imagefiles",
        nargs="+",
        help="The absolute or relative filenames where the relevant images can be found.",
    )
    imageops_parser.add_argument(
        "-p",
        "--particles",
        action="store_true",
        help=(
            "If present, the resulting mask will obscure everything except the bright particles "
            "on the fish in the given images. Otherwise the whole fish will be shown."
        ),
    )
    imageops_parser.add_argument(
        "-d",
        "--debug",
        action="count",
        default=1,
        help=(
            "Indicates intermediate processing images should be output for troubleshooting "
            "purposes. Including this argument once will yield one intermediate image per input "
            "file, twice will yield several intermediate images per input file."
        ),
    )
    imageops_parser.set_defaults(func=imageops_command)
    # endregion keyence

    # region dose_response parser
    # Create the parser for the dose_response script
    dose_response_parser = subparsers.add_parser(
        name="dose_response", help="Evaluate the dose response for different models"
    )
    dose_response_parser.add_argument(
        "filename",
        nargs="*",
        default=None,
        help="filenames containing dose response models",
    )
    dose_response_parser.set_defaults(func=dose_response_command)

    # endregion dose_response parser

    # region infection parser
    infection_parser = subparsers.add_parser(
        name="infection",
        help=(
            "Analyzer for images of whole zebrafish with stained neuromasts, for the "
            "purposes of measuring hair cell damage under drug-combination conditions. Reports "
            "values relative to control."
        ),
    )

    infection_parser.add_argument(
        "-cb",
        "--checkerboard",
        action="store_true",
        help=(
            "If present, the input will be treated as a checkerboard assay, with output produced "
            "accordingly."
        ),
    )

    infection_parser.add_argument(
        "-cv",
        "--conversions",
        default=[],
        nargs="*",
        type=infection._key_value_pair,
        help=(
            "List of conversions between dose concentration labels and concrete values, each as "
            "a separate argument, each delimited by an equals sign. For instance, ABC50 might be "
            "an abbreviation for the EC50 of drug ABC, in which case the concrete concentration "
            'can be supplied like "ABC50=ABC 1mM" (make sure to quote, or escape spaces).'
        ),
    )

    infection_parser.add_argument(
        "-ppc",
        "--plate-positive-control",
        default=[],
        nargs="*",
        help=(
            "Labels to treat as the positive control conditions in the plate schematic (i.e. "
            "conditions showing maximum effect). These wells are used to normalize all values in "
            "the plate for more interpretable results. Any number of values may be passed."
        ),
    )

    infection_parser.add_argument(
        "--plate-info",
        default=None,
        help=(
            "Any information identifying the plate(s) being analyzed that should be passed along "
            "to files created by this process."
        ),
    )

    infection_parser.add_argument(
        "-tp",
        "--treatment-platefile",
        help="CSV file containing a schematic of the plate in which the imaged fish were treated. "
        "Used to chart responses by treatment location, if desired. Row and column headers are "
        "optional. The cell values are essentially just arbitrary labels: results will be "
        "grouped and charted according to the supplied values.",
    )

    infection_parser.add_argument(
        "--absolute-chart",
        action="store_true",
        help=(
            "If present, a plate graphic will be generated with absolute (rather than relative) "
            "brightness values."
        ),
    )

    infection_parser.add_argument(
        "--talk",
        action="store_true",
        help=('If present, images will be generated with the Seaborn "talk" context.'),
    )
    set_arguments(infection_parser)
    utils.remove_arguments(infection_parser, "plate_ignore", "group_regex")
    infection_parser.set_defaults(func=infection_command)

    # endregion infection parser

    # region pipeline parser
    pipeline_parser = subparsers.add_parser(
        "pipeline",
        help=(
            "Analyzer for images of whole zebrafish with fluorescent neuromasts, for the "
            "purposes of measuring hair cell damage under drug-combination conditions. Reports "
            "values relative to control."
        ),
    )
    pipeline_parser.add_argument(
        "-cb",
        "--checkerboard",
        action="store_true",
        help=(
            "If present, the input will be treated as a checkerboard assay, with output produced "
            "accordingly."
        ),
    )

    pipeline_parser.add_argument(
        "-cv",
        "--conversions",
        default=[],
        nargs="*",
        type=pipeline._key_value_pair,
        help=(
            "List of conversions between dose concentration labels and concrete values, each as "
            "a separate argument, each delimited by an equals sign. For instance, ABC50 might be "
            "an abbreviation for the EC50 of drug ABC, in which case the concrete concentration "
            'can be supplied like "ABC50=ABC 1mM" (make sure to quote, or escape spaces).'
        ),
    )

    pipeline_parser.add_argument(
        "-ppc",
        "--plate-positive-control",
        default=[],
        nargs="*",
        help=(
            "Labels to treat as the positive control conditions in the plate schematic (i.e. "
            "conditions showing maximum effect). These wells are used to normalize all values in "
            "the plate for more interpretable results. Any number of values may be passed."
        ),
    )

    pipeline_parser.add_argument(
        "--plate-info",
        default=None,
        help=(
            "Any information identifying the plate(s) being analyzed that should be passed along "
            "to files created by this process."
        ),
    )

    pipeline_parser.add_argument(
        "-tp",
        "--treatment-platefile",
        help="CSV file containing a schematic of the plate in which the imaged fish were treated. "
        "Used to chart responses by treatment location, if desired. Row and column headers are "
        "optional. The cell values are essentially just arbitrary labels: results will be "
        "grouped and charted according to the supplied values.",
    )

    pipeline_parser.add_argument(
        "--absolute-chart",
        action="store_true",
        help=(
            "If present, a plate graphic will be generated with absolute (rather than relative) "
            "brightness values."
        ),
    )

    pipeline_parser.add_argument(
        "--talk",
        action="store_true",
        help=(
            'If present, images will be generated with the Seaborn "talk" context. Otherwise the '
            'default "notebook" context will be used. (See '
            "https://seaborn.pydata.org/generated/seaborn.set_context.html)"
        ),
    )
    set_arguments(pipeline_parser)
    pipeline_parser.set_defaults(func=pipeline_command)

    # endregion pipeline parser

    # region imagej scripts parser
    script_parser = subparsers.add_parser(
        "imagej-script", help="Write ImageJ scripts to provided directory"
    )
    script_parser.add_argument(
        "scripts",
        nargs="*",
        default=None,
        help="Names of scripts to write to 'directory', if not provided will save all scripts "
        "to directory.",
    )
    script_parser.add_argument(
        "-d",
        "--directory",
        default=".",
        help="Path (relative or absolute) to directory to write Imagej scripts, if not "
        "provided will save scripts to current directory",
        dest="directory",
    )
    script_parser.set_defaults(func=imagej_scripts_command)
    # endregion imagej scripts parser

    return top_parser


def pepita():
    # Create Parser
    parser = create_parser()
    # Parse arguments from stdin
    args = parser.parse_args()
    if (args.command != "config-file") and (args.command != "keyence-file"):
        # Read the config file
        read_config(args.config)
        # Replace default config with passed command line arguments
        merge_config_cli(args)
    # Call the desired subcommand
    args.func(args)


if __name__ == "__main__":
    pepita()



# Imports
# Standard Library Imports
import json
import os
import warnings

# External Imports
import numpy as np
import seaborn as sns

# Local Imports
from . import absolute, analyze, imageops, pipeline, utils
from .configuration import get_config_setting

# channel_main_infection = int(get_config_setting("channel_main_infection"))
# channel_subtr_infection = int(get_config_setting("channel_subtr_infection"))
# replacement_delim = get_config_setting("filename_replacement_delimiter")
# replacement_brfld = get_config_setting("filename_replacement_brightfield_infection").split(
#     replacement_delim
# )
# replacement_mask = get_config_setting("filename_replacement_mask_infection").split(
#     replacement_delim
# )
# replacement_subtr = get_config_setting("filename_replacement_subtr_infection").split(
#     replacement_delim
# )


class InfectionImage(analyze.Image):
    def __init__(self, filename, group, debug=0):
        super().__init__(filename, group, debug)
        self.channel = int(get_config_setting("channel_main_infection"))
        self.channel_subtr = int(get_config_setting("channel_subtr_infection"))
        self.particles = False
        replacement_delim = get_config_setting("filename_replacement_delimiter")
        self.replacement_brfld = get_config_setting(
            "filename_replacement_brightfield_infection"
        ).split(replacement_delim)
        self.replacement_mask = get_config_setting(
            "filename_replacement_mask_infection"
        ).split(replacement_delim)
        self.replacement_subtr = get_config_setting(
            "filename_replacement_subtr_infection"
        ).split(replacement_delim)

    def get_raw_value(self, threshold=0.02):
        if self.value is None:
            fl_img_masked = imageops.apply_mask(self.get_fl_img(), self.get_mask())
            max_value = imageops._get_bit_depth(fl_img_masked)[1]
            total = fl_img_masked.sum(
                dtype=np.uint64, where=(fl_img_masked > max_value * threshold)
            )
            self.value = total if total > 0 else np.nan
        return self.value


def log(results):
    return {key: np.log2(values).tolist() for key, values in results.items()}


def main(
    imagefiles,
    cap=-1,
    chartfile=None,
    checkerboard=False,
    conversions=None,
    debug=0,
    platefile=None,
    plate_control=None,
    plate_info=None,
    plate_positive_control=None,
    treatment_platefile=None,
    absolute_chart=False,
    silent=False,
    talk=False,
):
    if conversions is None:
        conversions = []
    if plate_control is None:
        plate_control = ["B"]
    if plate_positive_control is None:
        plate_positive_control = []
    hashfile = utils.get_inputs_hashfile(
        imagefiles=imagefiles, cap=cap, platefile=platefile, plate_control=plate_control
    )

    if talk:
        sns.set_context("talk")

    if debug == 0 and os.path.exists(hashfile):
        with open(hashfile, "r") as f:  # read cached results
            results = json.load(f)

        for group, relevant_values in results.items():
            if not silent:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", RuntimeWarning)
                    print(group, np.nanmedian(relevant_values), relevant_values)
    else:
        results = quantify_infection(
            imagefiles=imagefiles,
            cap=cap,
            debug=debug,
            platefile=platefile,
            plate_control=plate_control,
            silent=False,
        )
        with open(hashfile, "w") as f:  # cache results for reuse
            json.dump(results, f, ensure_ascii=False)

    if chartfile:
        analyze.chart(log(results), chartfile)

    conversions = dict(conversions)
    # Are these needed?
    # drug_conditions = _parse_results(results, conversions)
    # control_drugs = [
    #     utils.Cocktail(utils.Dose(control).drug) for control in plate_control
    # ]
    # models = {}

    results = {
        utils.Solution(key, conversions): value for key, value in results.items()
    }

    # generate plate schematics

    schematic = analyze.get_schematic(platefile, len(imagefiles), flat=False)

    max_result = np.log2(
        max(val for vals_list in results.values() for val in vals_list)
    )

    pipeline.generate_plate_schematic(
        schematic,
        log(results),
        conversions=conversions,
        plate_info=plate_info,
        well_count=96,
        cmap=sns.dark_palette("red", as_cmap=True),
        max_val=max_result,
    )


def quantify_infection(
    imagefiles, cap=-1, debug=0, platefile=None, plate_control=None, silent=False
):
    if plate_control is None:
        plate_control = ["B"]
    results = {}

    schematic = analyze.get_schematic(platefile, len(imagefiles))
    groups = list(dict.fromkeys(schematic))  # deduplicated copy of `schematic`

    images = [
        InfectionImage(filename, group, debug)
        for filename, group in zip(imagefiles, schematic)
    ]

    for group in groups:
        relevant_values = [
            absolute.get_absolute_value(img) for img in images if img.group == group
        ]
        results[group] = relevant_values
        if not silent:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", RuntimeWarning)
                print(group, np.nanmedian(relevant_values), relevant_values)

    return results


def _key_value_pair(argument, delimiter="="):
    return tuple(argument.split(delimiter))


def _parse_results(results, conversions):
    drug_conditions = {}
    for condition in results:
        solution = utils.Solution(condition, conversions)
        utils.put_multimap(drug_conditions, solution.get_cocktail(), solution)
    return drug_conditions



"""
Functions related to keyence
"""

# Imports
# Standard Library Imports
from __future__ import annotations
import csv
import importlib.resources
import xml.etree.ElementTree as element_tree

# External Imports

# Local Imports
from pepitatools import configuration

COLUMNS = ["B", "C", "D", "E", "F", "G"]
ROWS = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11]

LAYOUT_DEFAULT = [letter for letter in COLUMNS for _ in range(10)]

LENSES = None


def get_lenses() -> dict[str, dict[str, float]]:
    """Return the Lenses Configuration"""
    global LENSES  # Accessing module level variable
    if LENSES is None or len(LENSES) == 0:  # If LENSES has not been set, try to set it
        LENSES = {}  # Empty dictionary which will have the values read in
        keyence_file = configuration.get_config_setting("keyence_lenses_file")
        if keyence_file is None:  # No keyence file provided in config
            # No keyence_file set in the config, so read the default example file
            with importlib.resources.open_text(
                "pepitatools.data", "keyence_BZX800_lenses.csv"
            ) as f:
                # with open(data_file, encoding="utf8", newline="", mode="r") as f:
                reader = csv.reader(f, delimiter="\t")
                next(reader, None)  # Skip the header

                for name, working_distance, pixel_size, numerical_aperture in reader:
                    LENSES[name] = {
                        "Numerical Aperture": float(numerical_aperture),
                        "Pixel Size": float(pixel_size),
                        "Working Distance": float(working_distance),
                    }
    return LENSES


def write_example(filename):
    with importlib.resources.open_text(
        "pepitatools.data", "keyence_BZX800_lenses.csv"
    ) as keyence_file:
        example_keyence = keyence_file.read()
    with open(filename, "w") as example_file:
        example_file.write(example_keyence)


def extract_metadata(filename):
    with open(filename, "br") as f:
        file_content = f.read()
        start = file_content.find(b"<?xml")
        xml = str(file_content[start:], "utf-8")

    try:
        main = element_tree.fromstring(xml).find("SingleFileProperty")
    except element_tree.ParseError:
        print("No xml data found.")
        exit(1)

    metadata = {}

    metadata["Aperture"] = int(_getxml(main, "Shooting", "Parameter", "Aperture")) / 100
    metadata["Digital Zoom"] = int(_getxml(main, "Image", "DigitalZoom")) / 100
    metadata["Exposure"] = {}
    metadata["Exposure"]["Label"] = (
        _getxml(main, "Shooting", "Parameter", "ExposureTime", "Numerator")
        + "/"
        + _getxml(main, "Shooting", "Parameter", "ExposureTime", "Denominator")
        + "s"
    )
    metadata["Exposure"]["Value"] = int(
        _getxml(main, "Shooting", "Parameter", "ExposureTime", "Numerator")
    ) / int(_getxml(main, "Shooting", "Parameter", "ExposureTime", "Denominator"))
    metadata["Field of View"] = {}
    metadata["Field of View"]["X Start"] = int(
        _getxml(main, "Shooting", "StageLocationX")
    )
    metadata["Field of View"]["X End"] = int(
        _getxml(main, "Shooting", "StageLocationX")
    ) + int(_getxml(main, "Shooting", "XyStageRegion", "Width"))
    metadata["Field of View"]["Y Start"] = int(
        _getxml(main, "Shooting", "StageLocationY")
    )
    metadata["Field of View"]["Y End"] = int(
        _getxml(main, "Shooting", "StageLocationY")
    ) + int(_getxml(main, "Shooting", "XyStageRegion", "Height"))
    metadata["Field of View"]["Z"] = int(_getxml(main, "Shooting", "StageLocationZ"))
    metadata["Gain"] = {}
    metadata["Gain"]["Camera Gain"] = (
        int(_getxml(main, "Shooting", "Parameter", "CameraGain")) - 54
    )  # match readout in Keyence analysis software
    metadata["Gain"]["Camera Gain Unit"] = _getxml(
        main, "Shooting", "Parameter", "CameraGainUnit"
    )
    metadata["Gain"]["Camera Hardware Gain"] = int(
        _getxml(main, "Shooting", "Parameter", "CameraHardwareGain")
    )
    metadata["IsoSpeed"] = _getxml(main, "Shooting", "Parameter", "IsoSpeed")
    metadata["Lens"] = _getxml(main, "Lens", "LensName")
    metadata["Magnification"] = int(_getxml(main, "Lens", "Magnification")) / 100

    metadata["Numerical Aperture"] = get_lenses()[metadata["Lens"]][
        "Numerical Aperture"
    ]
    metadata["Pixel Size"] = get_lenses()[metadata["Lens"]]["Pixel Size"]
    metadata["Working Distance"] = get_lenses()[metadata["Lens"]]["Working Distance"]

    return metadata


def well_to_xy(well):
    col = well[0]
    row = int(well[1:])

    return COLUMNS.index(col) * 10 + ROWS.index(row) + 1


def xy_to_well(xy_num):
    xy_num -= 1

    return COLUMNS[xy_num // 10] + str(ROWS[xy_num % 10])


def _getxml(element, *fields):
    for field in fields:
        element = element.find(field)

    return element.text
